from pathlib import Path
from slither import Slither
from slither.slithir.operations import Assignment
from slither.slithir.variables import Constant

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_enum_max_min(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.19")
    slither = Slither(Path(TEST_DATA_DIR, "enum_max_min.sol").as_posix(), solc=solc_path)

    contract = slither.get_contract_from_name("D")[0]

    f = contract.get_function_from_full_name("a()")
    # TMP_1(uint256) := 2(uint256)
    assignment = f.slithir_operations[1]
    assert (
        isinstance(assignment, Assignment)
        and isinstance(assignment.rvalue, Constant)
        and assignment.rvalue.value == 2
    )

    f = contract.get_function_from_full_name("b()")
    # TMP_4(uint256) := 0(uint256)
    assignment = f.slithir_operations[1]
    assert (
        isinstance(assignment, Assignment)
        and isinstance(assignment.rvalue, Constant)
        and assignment.rvalue.value == 0
    )

    f = contract.get_function_from_full_name("c()")
    # TMP_7(uint256) := 1(uint256)
    assignment = f.slithir_operations[1]
    assert (
        isinstance(assignment, Assignment)
        and isinstance(assignment.rvalue, Constant)
        and assignment.rvalue.value == 1
    )

    f = contract.get_function_from_full_name("d()")
    # TMP_10(uint256) := 0(uint256)
    assignment = f.slithir_operations[1]
    assert (
        isinstance(assignment, Assignment)
        and isinstance(assignment.rvalue, Constant)
        and assignment.rvalue.value == 0
    )

    f = contract.get_function_from_full_name("e()")
    # TMP_13(uint256) := 0(uint256)
    assignment = f.slithir_operations[1]
    assert (
        isinstance(assignment, Assignment)
        and isinstance(assignment.rvalue, Constant)
        and assignment.rvalue.value == 0
    )

    f = contract.get_function_from_full_name("f()")
    # TMP_16(uint256) := 0(uint256)
    assignment = f.slithir_operations[1]
    assert (
        isinstance(assignment, Assignment)
        and isinstance(assignment.rvalue, Constant)
        and assignment.rvalue.value == 0
    )



"""
Nodes of the dominator tree
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from slither.core.cfg.node import Node


class DominatorNode:
    def __init__(self):
        self._succ: set[Node] = set()
        self._nodes: list[Node] = []

    def add_node(self, node: "Node"):
        self._nodes.append(node)

    def add_successor(self, succ: "Node"):
        self._succ.add(succ)

    @property
    def cfg_nodes(self) -> list["Node"]:
        return self._nodes

    @property
    def sucessors(self) -> set["Node"]:
        return self._succ



from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from slither.core.declarations import (
        Function,
        SolidityFunction,
        Contract,
        SolidityVariable,
    )
    from slither.core.variables.variable import Variable

### core.declaration

InternalCallType = Union["Function", "SolidityFunction"]
HighLevelCallType = tuple["Contract", Union["Function", "Variable"]]
LibraryCallType = tuple["Contract", "Function"]
LowLevelCallType = tuple[Union["Variable", "SolidityVariable"], str]



"""
Module printing summary of the contract
"""

from slither.core.declarations import Contract
from slither.printers.abstract_printer import AbstractPrinter
from slither.analyses.data_dependency.data_dependency import get_dependencies, SUPPORTED_TYPES
from slither.slithir.variables import TemporaryVariable, ReferenceVariable
from slither.utils.myprettytable import MyPrettyTable
from slither.utils.output import Output


def _get(v: SUPPORTED_TYPES, c: Contract) -> list[str]:
    return list(
        {
            d.name
            for d in get_dependencies(v, c)
            if not isinstance(d, (TemporaryVariable, ReferenceVariable)) and d.name
        }
    )


class DataDependency(AbstractPrinter):
    ARGUMENT = "data-dependency"
    HELP = "Print the data dependencies of the variables"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#data-dependencies"

    def output(self, _filename: str) -> Output:
        """
        _filename is not used
        Args:
            _filename(string)
        """

        all_tables = []
        all_txt = ""

        txt = ""
        for c in self.contracts:
            txt += f"\nContract {c.name}\n"
            table = MyPrettyTable(["Variable", "Dependencies"])
            for v in c.state_variables:
                assert v.name
                table.add_row([v.name, sorted(_get(v, c))])

            txt += str(table)

            txt += "\n"
            for f in c.functions_and_modifiers_declared:
                txt += f"\nFunction {f.full_name}\n"
                table = MyPrettyTable(["Variable", "Dependencies"])
                for v in f.variables:
                    table.add_row([v.name, sorted(_get(v, f))])
                for v in c.state_variables:
                    table.add_row([v.canonical_name, sorted(_get(v, f))])
                txt += str(table)
            self.info(txt)

            all_txt += txt
            all_tables.append((c.name, table))

        res = self.generate_output(all_txt)
        for name, table in all_tables:
            res.add_pretty_table(table, name)

        return res



from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.statements.pyth_unchecked import PythUnchecked


class PythUncheckedConfidence(PythUnchecked):
    """
    Documentation: This detector finds when the confidence level of a Pyth price is not checked
    """

    ARGUMENT = "pyth-unchecked-confidence"
    HELP = "Detect when the confidence level of a Pyth price is not checked"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#pyth-unchecked-confidence-level"
    WIKI_TITLE = "Pyth unchecked confidence level"
    WIKI_DESCRIPTION = "Detect when the confidence level of a Pyth price is not checked"
    WIKI_RECOMMENDATION = "Check the confidence level of a Pyth price. Visit https://docs.pyth.network/price-feeds/best-practices#confidence-intervals for more information."

    WIKI_EXPLOIT_SCENARIO = """
```solidity
import "@pythnetwork/pyth-sdk-solidity/IPyth.sol";
import "@pythnetwork/pyth-sdk-solidity/PythStructs.sol";

contract C {
    IPyth pyth;

    constructor(IPyth _pyth) {
        pyth = _pyth;
    }

    function bad(bytes32 id, uint256 age) public {
        PythStructs.Price memory price = pyth.getEmaPriceNoOlderThan(id, age);
        // Use price
    }
}
```
The function `A` uses the price without checking its confidence level.
"""

    PYTH_FUNCTIONS = [
        "getEmaPrice",
        "getEmaPriceNoOlderThan",
        "getEmaPriceUnsafe",
        "getPrice",
        "getPriceNoOlderThan",
        "getPriceUnsafe",
    ]

    PYTH_FIELD = "conf"



import logging
from pathlib import Path

from slither.utils.colors import green, yellow

logger = logging.getLogger("Slither")


def write_file(
    output_dir: Path,
    filename: str,
    content: str,
    allow_overwrite: bool = True,
    discard_if_exist: bool = False,
) -> None:
    """
    Write the content into output_dir/filename
    :param output_dir:
    :param filename:
    :param content:
    :param allow_overwrite: If true, allows to overwrite existing file (default: true). Emit warning if overwrites
    :param discard_if_exist: If true, it will not emit warning or overwrite the file if it exists, (default: False)
    :return:
    """
    file_to_write = Path(output_dir, filename)
    if file_to_write.exists():
        if discard_if_exist:
            return
        if not allow_overwrite:
            logger.info(yellow(f"{file_to_write} already exist and will not be overwritten"))
            return
        logger.info(yellow(f"Overwrite {file_to_write}"))
    else:
        logger.info(green(f"Write {file_to_write}"))
    with open(file_to_write, "w", encoding="utf8") as f:
        f.write(content)



from typing import Any, TypeVar
from collections.abc import Callable

from crytic_compile import CompilationUnit
from crytic_compile.source_unit import SourceUnit
from crytic_compile.utils.naming import Filename

from slither.core.declarations import Contract, Import, Pragma
from slither.core.declarations.custom_error_top_level import CustomErrorTopLevel
from slither.core.declarations.enum_top_level import EnumTopLevel
from slither.core.declarations.event_top_level import EventTopLevel
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.core.declarations.using_for_top_level import UsingForTopLevel
from slither.core.declarations.structure_top_level import StructureTopLevel
from slither.core.solidity_types import TypeAlias
from slither.core.variables.top_level_variable import TopLevelVariable
from slither.slithir.variables import Constant


def _dict_contain(d1: dict, d2: dict) -> bool:
    """
    Return true if d1 is included in d2
    """
    d2_keys = d2.keys()
    return all(item in d2_keys for item in d1)


class FileScope:
    def __init__(self, filename: Filename) -> None:
        self.filename = filename
        self.accessible_scopes: list[FileScope] = []
        self.exported_symbols: set[int] = set()

        self.contracts: dict[str, Contract] = {}
        # Custom error are a list instead of a dict
        # Because we parse the function signature later on
        # So we simplify the logic and have the scope fields all populated
        self.custom_errors: set[CustomErrorTopLevel] = set()
        self.enums: dict[str, EnumTopLevel] = {}
        # Functions is a list instead of a dict
        # Because we parse the function signature later on
        # So we simplify the logic and have the scope fields all populated
        self.functions: set[FunctionTopLevel] = set()
        self.events: set[EventTopLevel] = set()
        self.using_for_directives: set[UsingForTopLevel] = set()
        self.imports: set[Import] = set()
        self.pragmas: set[Pragma] = set()
        self.structures: dict[str, StructureTopLevel] = {}
        self.variables: dict[str, TopLevelVariable] = {}

        # Renamed created by import
        # import A as B
        # local name -> original name (A -> B)
        self.renaming: dict[str, str] = {}

        # User defined types
        # Name -> type alias
        self.type_aliases: dict[str, TypeAlias] = {}

    def add_accessible_scopes(self) -> bool:
        """
        Add information from accessible scopes. Return true if new information was obtained
        :return:
        :rtype:
        """

        learn_something = False
        for new_scope in self.accessible_scopes:
            # To support using for directives on user defined types and user defined functions,
            # we need to propagate the using for directives from the imported file to the importing file
            # since it is not reflected in the "exportedSymbols" field of the AST.
            if not new_scope.using_for_directives.issubset(self.using_for_directives):
                self.using_for_directives |= new_scope.using_for_directives
                learn_something = True
            if not _dict_contain(new_scope.type_aliases, self.type_aliases):
                self.type_aliases.update(new_scope.type_aliases)
                learn_something = True
            if not new_scope.functions.issubset(self.functions):
                self.functions |= new_scope.functions
                learn_something = True

            # To get around this bug for aliases https://github.com/ethereum/solidity/pull/11881,
            # we propagate the exported_symbols from the imported file to the importing file
            # See tests/e2e/solc_parsing/test_data/top-level-nested-import-0.7.1.sol
            if not new_scope.exported_symbols.issubset(self.exported_symbols):
                self.exported_symbols |= new_scope.exported_symbols
                learn_something = True

            # This is need to support aliasing when we do a late lookup using SolidityImportPlaceholder
            if not _dict_contain(new_scope.renaming, self.renaming):
                self.renaming.update(new_scope.renaming)
                learn_something = True

        return learn_something

    def get_contract_from_name(self, name: str | Constant) -> Contract | None:
        if isinstance(name, Constant):
            return self.contracts.get(name.name, None)
        return self.contracts.get(name, None)

    AbstractReturnType = TypeVar("AbstractReturnType")

    def _generic_source_unit_getter(
        self,
        crytic_compile_compilation_unit: CompilationUnit,
        name: str,
        getter: Callable[[SourceUnit], dict[str, AbstractReturnType]],
    ) -> AbstractReturnType | None:
        assert self.filename in crytic_compile_compilation_unit.source_units

        source_unit = crytic_compile_compilation_unit.source_unit(self.filename)

        if name in getter(source_unit):
            return getter(source_unit)[name]

        for scope in self.accessible_scopes:
            source_unit = crytic_compile_compilation_unit.source_unit(scope.filename)
            if name in getter(source_unit):
                return getter(source_unit)[name]

        return None

    def bytecode_init(
        self, crytic_compile_compilation_unit: CompilationUnit, contract_name: str
    ) -> str | None:
        """
        Return the init bytecode

        Args:
            crytic_compile_compilation_unit:
            contract_name:

        Returns:

        """
        getter: Callable[[SourceUnit], dict[str, str]] = lambda x: x.bytecodes_init
        return self._generic_source_unit_getter(
            crytic_compile_compilation_unit, contract_name, getter
        )

    def bytecode_runtime(
        self, crytic_compile_compilation_unit: CompilationUnit, contract_name: str
    ) -> str | None:
        """
        Return the runtime bytecode

        Args:
            crytic_compile_compilation_unit:
            contract_name:

        Returns:

        """
        getter: Callable[[SourceUnit], dict[str, str]] = lambda x: x.bytecodes_runtime
        return self._generic_source_unit_getter(
            crytic_compile_compilation_unit, contract_name, getter
        )

    def srcmap_init(
        self, crytic_compile_compilation_unit: CompilationUnit, contract_name: str
    ) -> list[str] | None:
        """
        Return the init scrmap

        Args:
            crytic_compile_compilation_unit:
            contract_name:

        Returns:

        """
        getter: Callable[[SourceUnit], dict[str, list[str]]] = lambda x: x.srcmaps_init
        return self._generic_source_unit_getter(
            crytic_compile_compilation_unit, contract_name, getter
        )

    def srcmap_runtime(
        self, crytic_compile_compilation_unit: CompilationUnit, contract_name: str
    ) -> list[str] | None:
        """
        Return the runtime srcmap

        Args:
            crytic_compile_compilation_unit:
            contract_name:

        Returns:

        """
        getter: Callable[[SourceUnit], dict[str, list[str]]] = lambda x: x.srcmaps_runtime
        return self._generic_source_unit_getter(
            crytic_compile_compilation_unit, contract_name, getter
        )

    def abi(self, crytic_compile_compilation_unit: CompilationUnit, contract_name: str) -> Any:
        """
        Return the abi

        Args:
            crytic_compile_compilation_unit:
            contract_name:

        Returns:

        """
        getter: Callable[[SourceUnit], dict[str, list[str]]] = lambda x: x.abis
        return self._generic_source_unit_getter(
            crytic_compile_compilation_unit, contract_name, getter
        )

    # region Built in definitions
    ###################################################################################
    ###################################################################################

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, str):
            return other == self.filename
        return NotImplemented

    def __neq__(self, other: Any) -> bool:
        if isinstance(other, str):
            return other != self.filename
        return NotImplemented

    def __str__(self) -> str:
        return str(self.filename.relative)

    def __hash__(self) -> int:
        return hash(self.filename.relative)

    # endregion



"""
Module printing summary of the contract
"""

import logging

from slither.core.declarations import SolidityFunction, Function
from slither.core.variables.state_variable import StateVariable
from slither.printers.abstract_printer import AbstractPrinter
from slither.printers.summary.loc import compute_loc_metrics
from slither.slithir.operations import (
    LowLevelCall,
    HighLevelCall,
    Transfer,
    Send,
    SolidityCall,
)
from slither.utils import output
from slither.utils.code_complexity import compute_cyclomatic_complexity
from slither.utils.colors import green, red, yellow
from slither.utils.myprettytable import MyPrettyTable
from slither.utils.standard_libraries import is_standard_library
from slither.core.cfg.node import NodeType


class PrinterHumanSummary(AbstractPrinter):
    ARGUMENT = "human-summary"
    HELP = "Print a human-readable summary of the contracts"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#human-summary"

    @staticmethod
    def _get_summary_erc20(contract):
        functions_name = [f.name for f in contract.functions]
        state_variables = [v.name for v in contract.state_variables]

        pause = "pause" in functions_name

        if "mint" in functions_name:
            if "mintingFinished" in state_variables:
                mint_unlimited = False
            else:
                mint_unlimited = True
        else:
            mint_unlimited = None  # no minting

        race_condition_mitigated = (
            "increaseApproval" in functions_name or "safeIncreaseAllowance" in functions_name
        )

        return pause, mint_unlimited, race_condition_mitigated

    def get_summary_erc20(self, contract):
        txt = ""

        pause, mint_unlimited, race_condition_mitigated = self._get_summary_erc20(contract)

        if pause:
            txt += yellow("Pausable") + "\n"

        if mint_unlimited is None:
            txt += green("No Minting") + "\n"
        else:
            if mint_unlimited:
                txt += red("∞ Minting") + "\n"
            else:
                txt += yellow("Minting") + "\n"

        if not race_condition_mitigated:
            txt += red("Approve Race Cond.") + "\n"

        return txt

    def _get_detectors_result(self) -> tuple[list[dict], int, int, int, int, int]:
        # disable detectors logger
        logger = logging.getLogger("Detectors")
        logger.setLevel(logging.ERROR)

        checks_optimization = self.slither.detectors_optimization
        checks_informational = self.slither.detectors_informational
        checks_low = self.slither.detectors_low
        checks_medium = self.slither.detectors_medium
        checks_high = self.slither.detectors_high

        issues_optimization = [c.detect() for c in checks_optimization]
        issues_optimization = [c for c in issues_optimization if c]
        issues_optimization = [item for sublist in issues_optimization for item in sublist]

        issues_informational = [c.detect() for c in checks_informational]
        issues_informational = [c for c in issues_informational if c]
        issues_informational = [item for sublist in issues_informational for item in sublist]

        issues_low = [c.detect() for c in checks_low]
        issues_low = [c for c in issues_low if c]
        issues_low = [item for sublist in issues_low for item in sublist]

        issues_medium = (c.detect() for c in checks_medium)
        issues_medium = [c for c in issues_medium if c]
        issues_medium = [item for sublist in issues_medium for item in sublist]

        issues_high = [c.detect() for c in checks_high]
        issues_high = [c for c in issues_high if c]
        issues_high = [item for sublist in issues_high for item in sublist]

        all_results = (
            issues_optimization + issues_informational + issues_low + issues_medium + issues_high
        )

        return (
            all_results,
            len(issues_optimization),
            len(issues_informational),
            len(issues_low),
            len(issues_medium),
            len(issues_high),
        )

    def get_detectors_result(self) -> tuple[str, list[dict], int, int, int, int, int]:
        (
            all_results,
            optimization,
            informational,
            low,
            medium,
            high,
        ) = self._get_detectors_result()
        txt = f"Number of optimization issues: {green(optimization)}\n"
        txt += f"Number of informational issues: {green(informational)}\n"
        txt += f"Number of low issues: {green(low)}\n"
        if medium > 0:
            txt += f"Number of medium issues: {yellow(medium)}\n"
        else:
            txt += f"Number of medium issues: {green(medium)}\n"
        if high > 0:
            txt += f"Number of high issues: {red(high)}\n"
        else:
            txt += f"Number of high issues: {green(high)}\n\n"

        return txt, all_results, optimization, informational, low, medium, high

    @staticmethod
    def _is_complex_code(contract):
        for f in contract.functions:
            if compute_cyclomatic_complexity(f) > 7:
                return True
        return False

    def is_complex_code(self, contract):
        """
            Check if the code is complex
            Heuristic, the code is complex if:
                - One function has a cyclomatic complexity > 7
        Args:
            contract
        """

        is_complex = self._is_complex_code(contract)

        result = red("Yes") if is_complex else green("No")
        return result

    @staticmethod
    def _number_functions(contract):
        return len(contract.functions)

    def _get_number_of_assembly_lines(self) -> int:
        total_asm_lines = 0
        for contract in self.contracts:
            for function in contract.functions_declared:
                for node in function.nodes:
                    if node.type == NodeType.ASSEMBLY:
                        inline_asm = node.inline_asm
                        if inline_asm:
                            total_asm_lines += len(inline_asm.splitlines())
        return total_asm_lines

    def _compilation_type(self):
        if self.slither.crytic_compile is None:
            return "Compilation non standard\n"
        return f"Compiled with {self.slither.crytic_compile.type!s}\n"

    def _number_contracts(self) -> tuple[int, int, int]:
        contracts = self.slither.contracts
        deps = [c for c in contracts if c.is_from_dependency()]
        tests = [c for c in contracts if c.is_test]
        return len(contracts) - len(deps) - len(tests), len(deps), len(tests)

    def _standard_libraries(self):
        libraries = []
        for contract in self.contracts:
            lib = is_standard_library(contract)
            if lib:
                libraries.append(lib)

        return libraries

    def _ercs(self):
        ercs = []
        for contract in self.contracts:
            ercs += contract.ercs()
        return list(set(ercs))

    def _get_features(self, contract):
        has_payable = False
        can_send_eth = False
        can_selfdestruct = False
        has_ecrecover = False
        can_delegatecall = False
        has_token_interaction = False

        has_assembly = False

        use_abi_encoder = False

        for compilation_unit in self.slither.compilation_units:
            for pragma in compilation_unit.pragma_directives:
                if (
                    pragma.source_mapping.filename.absolute
                    == contract.source_mapping.filename.absolute
                ):
                    if pragma.is_abi_encoder_v2:
                        use_abi_encoder = True

        for function in contract.functions:
            if function.payable:
                has_payable = True

            if function.contains_assembly:
                has_assembly = True

            for ir in function.slithir_operations:
                if isinstance(ir, (LowLevelCall, HighLevelCall, Send, Transfer)) and ir.call_value:
                    can_send_eth = True
                if isinstance(ir, SolidityCall) and ir.function in [
                    SolidityFunction("suicide(address)"),
                    SolidityFunction("selfdestruct(address)"),
                ]:
                    can_selfdestruct = True
                if isinstance(ir, SolidityCall) and ir.function == SolidityFunction(
                    "ecrecover(bytes32,uint8,bytes32,bytes32)"
                ):
                    has_ecrecover = True
                if isinstance(ir, LowLevelCall) and ir.function_name in [
                    "delegatecall",
                    "callcode",
                ]:
                    can_delegatecall = True
                if isinstance(ir, HighLevelCall):
                    if (
                        isinstance(ir.function, (Function, StateVariable))
                        and ir.function.contract.is_possible_token
                    ):
                        has_token_interaction = True

        return {
            "Receive ETH": has_payable,
            "Send ETH": can_send_eth,
            "Selfdestruct": can_selfdestruct,
            "Ecrecover": has_ecrecover,
            "Delegatecall": can_delegatecall,
            "Tokens interaction": has_token_interaction,
            "AbiEncoderV2": use_abi_encoder,
            "Assembly": has_assembly,
            "Upgradeable": contract.is_upgradeable,
            "Proxy": contract.is_upgradeable_proxy,
        }

    def _get_contracts(self, txt: str) -> str:
        (
            number_contracts,
            number_contracts_deps,
            number_contracts_tests,
        ) = self._number_contracts()
        txt += f"Total number of contracts in source files: {number_contracts}\n"
        if number_contracts_deps > 0:
            txt += f"Number of contracts in dependencies: {number_contracts_deps}\n"
        if number_contracts_tests > 0:
            txt += f"Number of contracts in tests       : {number_contracts_tests}\n"
        return txt

    def _get_number_lines(self, txt: str, results: dict) -> tuple[str, dict]:
        loc = compute_loc_metrics(self.slither)
        txt += "Source lines of code (SLOC) in source files: "
        txt += f"{loc.src.sloc}\n"
        if loc.dep.sloc > 0:
            txt += "Source lines of code (SLOC) in dependencies: "
            txt += f"{loc.dep.sloc}\n"
        if loc.test.sloc > 0:
            txt += "Source lines of code (SLOC) in tests       : "
            txt += f"{loc.test.sloc}\n"
        results["number_lines"] = loc.src.sloc
        results["number_lines__dependencies"] = loc.dep.sloc
        total_asm_lines = self._get_number_of_assembly_lines()
        txt += f"Number of  assembly lines: {total_asm_lines}\n"
        results["number_lines_assembly"] = total_asm_lines
        return txt, results

    def output(self, _filename):
        """
        _filename is not used
            Args:
                _filename(string)
        """

        txt = "\n"
        txt += self._compilation_type()

        results = {
            "contracts": {"elements": []},
            "number_lines": 0,
            "number_lines_in_dependencies": 0,
            "number_lines_assembly": 0,
            "standard_libraries": [],
            "ercs": [],
            "number_findings": {},
            "detectors": [],
        }
        txt = self._get_contracts(txt)
        txt, results = self._get_number_lines(txt, results)
        (
            txt_detectors,
            detectors_results,
            optimization,
            info,
            low,
            medium,
            high,
        ) = self.get_detectors_result()
        txt += txt_detectors

        results["number_findings"] = {
            "optimization_issues": optimization,
            "informational_issues": info,
            "low_issues": low,
            "medium_issues": medium,
            "high_issues": high,
        }
        results["detectors"] = detectors_results

        libs = self._standard_libraries()
        if libs:
            txt += f"\nUse: {', '.join(libs)}\n"
            results["standard_libraries"] = [str(lib) for lib in libs]

        ercs = self._ercs()
        if ercs:
            txt += f"ERCs: {', '.join(ercs)}\n"
            results["ercs"] = [str(e) for e in ercs]

        table = MyPrettyTable(
            ["Name", "# functions", "ERCS", "ERC20 info", "Complex code", "Features"]
        )
        for contract in self.slither.contracts_derived:
            if contract.is_from_dependency() or contract.is_test:
                continue

            is_complex = self.is_complex_code(contract)
            number_functions = self._number_functions(contract)
            ercs = ",".join(contract.ercs())
            is_erc20 = contract.is_erc20()
            erc20_info = ""
            if is_erc20:
                erc20_info += self.get_summary_erc20(contract)

            features = "\n".join(
                [name for name, to_print in self._get_features(contract).items() if to_print]
            )

            table.add_row(
                [
                    contract.name,
                    number_functions,
                    ercs,
                    erc20_info,
                    is_complex,
                    features,
                ]
            )

        self.info(txt + "\n" + str(table))

        results_contract = output.Output("")
        for contract in self.slither.contracts_derived:
            if contract.is_test or contract.is_from_dependency():
                continue

            contract_d = {
                "contract_name": contract.name,
                "is_complex_code": self._is_complex_code(contract),
                "is_erc20": contract.is_erc20(),
                "number_functions": self._number_functions(contract),
                "features": [
                    name for name, to_print in self._get_features(contract).items() if to_print
                ],
            }
            if contract_d["is_erc20"]:
                pause, mint_limited, race_condition_mitigated = self._get_summary_erc20(contract)
                contract_d["erc20_pause"] = pause
                if mint_limited is not None:
                    contract_d["erc20_can_mint"] = True
                    contract_d["erc20_mint_limited"] = mint_limited
                else:
                    contract_d["erc20_can_mint"] = False
                contract_d["erc20_race_condition_mitigated"] = race_condition_mitigated
            results_contract.add_contract(contract, additional_fields=contract_d)

        results["contracts"]["elements"] = results_contract.elements

        json = self.generate_output(txt, additional_fields=results)

        return json



"""
This class is used for the SSA version of slithIR
It is similar to the non-SSA version of slithIR
as the TemporaryVariable are in SSA form in both version
"""

from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.reference import ReferenceVariable
from slither.slithir.variables.tuple import TupleVariable


class TemporaryVariableSSA(TemporaryVariable):
    def __init__(self, temporary: TemporaryVariable) -> None:
        super().__init__(temporary.node, temporary.index)

        self._non_ssa_version = temporary

    @property
    def non_ssa_version(self) -> TemporaryVariable | TupleVariable | ReferenceVariable:
        return self._non_ssa_version



from Crypto.Hash import keccak


def get_function_id(sig: str) -> int:
    """'
        Return the function id of the given signature
    Args:
        sig (str)
    Return:
        (int)
    """
    digest = keccak.new(digest_bits=256)
    digest.update(sig.encode("utf8"))
    return int("0x" + digest.hexdigest()[:8], 16)


def get_event_id(sig: str) -> int:
    """'
        Return the event id of the given signature
    Args:
        sig (str)
    Return:
        (int)
    """
    digest = keccak.new(digest_bits=256)
    digest.update(sig.encode("utf8"))
    return int("0x" + digest.hexdigest(), 16)



from typing import TYPE_CHECKING, Optional

from slither.core.declarations import Event, Enum, Structure
from slither.core.declarations.contract import Contract
from slither.core.declarations.custom_error import CustomError
from slither.core.declarations.custom_error_contract import CustomErrorContract
from slither.core.declarations.custom_error_top_level import CustomErrorTopLevel
from slither.core.declarations.function import Function
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.core.declarations.solidity_import_placeholder import SolidityImportPlaceHolder
from slither.core.declarations.solidity_variables import (
    SOLIDITY_FUNCTIONS,
    SOLIDITY_VARIABLES,
    SolidityFunction,
    SolidityVariable,
)
from slither.core.scope.scope import FileScope
from slither.core.solidity_types import (
    ArrayType,
    FunctionType,
    MappingType,
    TypeAlias,
)
from slither.core.variables.top_level_variable import TopLevelVariable
from slither.core.variables.variable import Variable
from slither.exceptions import SlitherError
from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.exceptions import VariableNotFound

if TYPE_CHECKING:
    from slither.solc_parsing.declarations.function import FunctionSolc


# CallerContext =Union["ContractSolc", "FunctionSolc", "CustomErrorSolc", "StructureTopLevelSolc"]


def _get_pointer_name(variable: Variable) -> str | None:
    curr_type = variable.type
    while isinstance(curr_type, (ArrayType, MappingType)):
        if isinstance(curr_type, ArrayType):
            curr_type = curr_type.type
        else:
            assert isinstance(curr_type, MappingType)
            curr_type = curr_type.type_to

    if isinstance(curr_type, FunctionType):
        return variable.name + curr_type.parameters_signature
    return None


def _find_variable_from_ref_declaration(
    referenced_declaration: int | None,
    all_contracts: list["Contract"],
    all_functions: list["Function"],
    function_parser: Optional["FunctionSolc"],
    contract_declarer: Optional["Contract"],
) -> Contract | Function | None:
    """
    Reference declarations take the highest priority, but they are not available for legacy AST.
    """
    if referenced_declaration is None:
        return None
    # We look for variable declared with the referencedDeclaration attribute
    if function_parser is not None and referenced_declaration in function_parser.variables_renamed:
        return function_parser.variables_renamed[referenced_declaration].underlying_variable

    if (
        contract_declarer is not None
        and referenced_declaration in contract_declarer.state_variables_by_ref_id
    ):
        return contract_declarer.state_variables_by_ref_id[referenced_declaration]
    # Ccontracts  ids are the referenced declaration
    # This is not true for the functions, as we dont always have the referenced_declaration
    # But maybe we could? (TODO)
    for contract_candidate in all_contracts:
        if contract_candidate and contract_candidate.id == referenced_declaration:
            return contract_candidate
    for function_candidate in all_functions:
        if function_candidate.id == referenced_declaration and not function_candidate.is_shadowed:
            return function_candidate
    return None


def _find_variable_in_function_parser(
    var_name: str,
    function_parser: Optional["FunctionSolc"],
) -> Variable | None:
    if function_parser is None:
        return None
    # If not found, check for name
    func_variables = function_parser.underlying_function.variables_as_dict
    if var_name in func_variables:
        return func_variables[var_name]
    # A local variable can be a pointer
    # for example
    # function test(function(uint) internal returns(bool) t) interna{
    # Will have a local variable t which will match the signature
    # t(uint256)
    func_variables_ptr = {
        _get_pointer_name(f): f for f in function_parser.underlying_function.variables
    }
    if var_name and var_name in func_variables_ptr:
        return func_variables_ptr[var_name]

    return None


def find_top_level(
    var_name: str, scope: "FileScope"
) -> tuple[
    Enum | Structure | SolidityImportPlaceHolder | CustomError | TopLevelVariable | None, bool
]:
    """
    Return the top level variable use, and a boolean indicating if the variable returning was cretead
    If the variable was created, it has no source_mapping

    :param var_name:
    :type var_name:
    :param sl:
    :type sl:
    :return:
    :rtype:
    """
    if var_name in scope.type_aliases:
        return scope.type_aliases[var_name], False

    if var_name in scope.structures:
        return scope.structures[var_name], False

    if var_name in scope.enums:
        return scope.enums[var_name], False

    for event in scope.events:
        if var_name == event.full_name:
            return event, False

    for import_directive in scope.imports:
        if import_directive.alias == var_name:
            new_val = SolidityImportPlaceHolder(import_directive)
            return new_val, True

    if var_name in scope.variables:
        return scope.variables[var_name], False

    # This path should be reached only after the top level custom error have been parsed
    # If not, slither will crash
    # It does not seem to be reacheable, but if so, we will have to adapt the order of logic
    # This must be at the end, because other top level objects might require to go over "_find_top_level"
    # Before the parsing of the top level custom error
    # For example, a top variable that use another top level variable
    # IF more top level objects are added to Solidity, we have to be careful with the order of the lookup
    # in this function
    try:
        for custom_error in scope.custom_errors:
            if custom_error.solidity_signature == var_name:
                return custom_error, False
    except ValueError:
        # This can happen as custom error sol signature might not have been built
        # when find_variable was called
        # TODO refactor find_variable to prevent this from happening
        pass

    return None, False


def _find_in_contract(
    var_name: str,
    contract: Contract | None,
    contract_declarer: Contract | None,
    is_super: bool,
    is_identifier_path: bool = False,
) -> Variable | Function | Contract | Event | Enum | Structure | CustomError | None:
    if contract is None or contract_declarer is None:
        return None

    # variable are looked from the contract declarer
    contract_variables = contract_declarer.variables_as_dict
    if var_name in contract_variables:
        return contract_variables[var_name]

    # A state variable can be a pointer
    conc_variables_ptr = {_get_pointer_name(f): f for f in contract_declarer.variables}
    if var_name and var_name in conc_variables_ptr:
        return conc_variables_ptr[var_name]

    if is_super:
        getter_available = lambda f: f.functions_declared
        d = {f.canonical_name: f for f in contract.functions}
        functions = {
            f.full_name: f
            for f in contract_declarer.available_elements_from_inheritances(
                d, getter_available
            ).values()
        }
    else:
        functions = {f.full_name: f for f in contract.functions if not f.is_shadowed}
    if var_name in functions:
        return functions[var_name]

    if is_super:
        getter_available = lambda m: m.modifiers_declared
        d = {m.canonical_name: m for m in contract.modifiers}
        modifiers = {
            m.full_name: m
            for m in contract_declarer.available_elements_from_inheritances(
                d, getter_available
            ).values()
        }
    else:
        modifiers = contract.available_modifiers_as_dict()
    if var_name in modifiers:
        return modifiers[var_name]

    if is_identifier_path:
        for sig, modifier in modifiers.items():
            if "(" in sig:
                sig = sig[0 : sig.find("(")]
                if sig == var_name:
                    return modifier

    type_aliases = contract.type_aliases_as_dict
    if var_name in type_aliases:
        return type_aliases[var_name]

    # structures are looked on the contract declarer
    structures = contract.structures_as_dict
    if var_name in structures:
        return structures[var_name]

    events = contract.events_as_dict
    if var_name in events:
        return events[var_name]

    enums = contract.enums_as_dict
    if var_name in enums:
        return enums[var_name]

    # Note: contract.custom_errors_as_dict uses the name (not the sol sig) as key
    # This is because when the dic is populated the underlying object is not yet parsed
    # As a result, we need to iterate over all the custom errors here instead of using the dict
    custom_errors = contract.custom_errors
    try:
        for custom_error in custom_errors:
            if var_name in [custom_error.solidity_signature, custom_error.full_name]:
                return custom_error
    except ValueError:
        # This can happen as custom error sol signature might not have been built
        # when find_variable was called
        # TODO refactor find_variable to prevent this from happening
        pass

    # If the enum is referred as its name rather than its canonicalName
    enums = {e.name: e for e in contract.enums}
    if var_name in enums:
        return enums[var_name]

    return None


def _find_variable_init(
    caller_context: CallerContextExpression,
) -> tuple[
    list[Contract],
    list["Function"],
    FileScope,
]:
    from slither.solc_parsing.declarations.contract import ContractSolc
    from slither.solc_parsing.declarations.function import FunctionSolc
    from slither.solc_parsing.declarations.structure_top_level import StructureTopLevelSolc
    from slither.solc_parsing.variables.top_level_variable import TopLevelVariableSolc
    from slither.solc_parsing.declarations.event_top_level import EventTopLevelSolc
    from slither.solc_parsing.declarations.custom_error import CustomErrorSolc

    direct_contracts: list[Contract]
    direct_functions_parser: list[Function]
    scope: FileScope

    if isinstance(caller_context, FileScope):
        direct_contracts = []
        direct_functions_parser = []
        scope = caller_context
    elif isinstance(caller_context, ContractSolc):
        direct_contracts = [caller_context.underlying_contract]
        direct_functions_parser = [
            f.underlying_function
            for f in caller_context.functions_parser + caller_context.modifiers_parser
        ]
        scope = caller_context.underlying_contract.file_scope
    elif isinstance(caller_context, FunctionSolc):
        if caller_context.contract_parser:
            direct_contracts = [caller_context.contract_parser.underlying_contract]
            direct_functions_parser = [
                f.underlying_function
                for f in caller_context.contract_parser.functions_parser
                + caller_context.contract_parser.modifiers_parser
            ]
        else:
            # Top level functions
            direct_contracts = []
            direct_functions_parser = []
        underlying_function = caller_context.underlying_function
        if isinstance(underlying_function, FunctionTopLevel):
            scope = underlying_function.file_scope
        else:
            assert isinstance(underlying_function, FunctionContract)
            scope = underlying_function.contract_declarer.file_scope

    elif isinstance(caller_context, StructureTopLevelSolc):
        direct_contracts = []
        direct_functions_parser = []
        scope = caller_context.underlying_structure.file_scope
    elif isinstance(caller_context, TopLevelVariableSolc):
        direct_contracts = []
        direct_functions_parser = []
        scope = caller_context.underlying_variable.file_scope
    elif isinstance(caller_context, EventTopLevelSolc):
        direct_contracts = []
        direct_functions_parser = []
        scope = caller_context.underlying_event.file_scope
    elif isinstance(caller_context, CustomErrorSolc):
        if caller_context.contract_parser:
            direct_contracts = [caller_context.contract_parser.underlying_contract]
            direct_functions_parser = [
                f.underlying_function
                for f in caller_context.contract_parser.functions_parser
                + caller_context.contract_parser.modifiers_parser
            ]
        else:
            # Top level custom error
            direct_contracts = []
            direct_functions_parser = []
        underlying_custom_error = caller_context.underlying_custom_error
        if isinstance(underlying_custom_error, CustomErrorTopLevel):
            scope = underlying_custom_error.file_scope
        else:
            assert isinstance(underlying_custom_error, CustomErrorContract)
            scope = underlying_custom_error.contract.file_scope
    else:
        raise SlitherError(
            f"{type(caller_context)} ({caller_context} is not valid for find_variable"
        )

    return direct_contracts, direct_functions_parser, scope


def find_variable(
    var_name: str,
    caller_context: CallerContextExpression,
    referenced_declaration: int | None = None,
    is_super: bool = False,
    is_identifier_path: bool = False,
) -> tuple[
    Variable
    | Function
    | Contract
    | SolidityVariable
    | SolidityFunction
    | Event
    | Enum
    | Structure
    | CustomError
    | TypeAlias,
    bool,
]:
    """
    Return the variable found and a boolean indicating if the variable was created
    If the variable was created, it has no source mapping, and it the caller must add it

    :param var_name:
    :type var_name:
    :param caller_context:
    :type caller_context:
    :param referenced_declaration:
    :type referenced_declaration:
    :param is_super:
    :type is_super:
    :param is_identifier_path:
    :type is_identifier_path:
    :return:
    :rtype:
    """
    from slither.solc_parsing.declarations.function import FunctionSolc
    from slither.solc_parsing.declarations.contract import ContractSolc
    from slither.solc_parsing.declarations.custom_error import CustomErrorSolc

    # variable are looked from the contract declarer
    # functions can be shadowed, but are looked from the contract instance, rather than the contract declarer
    # the difference between function and variable come from the fact that an internal call, or an variable access
    # in a function does not behave similariy, for example in:
    # contract C{
    #   function f(){
    #     state_var = 1
    #     f2()
    #  }
    # state_var will refer to C.state_var, no mater if C is inherited
    # while f2() will refer to the function definition of the inherited contract (C.f2() in the context of C, or
    # the contract inheriting from C)
    # for events it's unclear what should be the behavior, as they can be shadowed, but there is not impact
    # structure/enums cannot be shadowed

    direct_contracts, direct_functions, current_scope = _find_variable_init(caller_context)
    # Only look for reference declaration in the direct contract, see comment at the end
    # Reference looked are split between direct and all
    # Because functions are copied between contracts, two functions can have the same ref
    # So we need to first look with respect to the direct context

    if var_name in current_scope.renaming:
        var_name = current_scope.renaming[var_name]

    contract: Contract | None = None
    contract_declarer: Contract | None = None
    if isinstance(caller_context, ContractSolc):
        contract = caller_context.underlying_contract
        contract_declarer = caller_context.underlying_contract
    elif isinstance(caller_context, FunctionSolc):
        underlying_func = caller_context.underlying_function
        if isinstance(underlying_func, FunctionContract):
            contract = underlying_func.contract
            contract_declarer = underlying_func.contract_declarer
        else:
            assert isinstance(underlying_func, FunctionTopLevel)
    elif isinstance(caller_context, CustomErrorSolc):
        underlying_custom_error = caller_context.underlying_custom_error
        if isinstance(underlying_custom_error, CustomErrorContract):
            contract = underlying_custom_error.contract
            # We check for contract variables here because _find_in_contract
            # will return since in this case the contract_declarer is None
            for var in contract.variables:
                if var_name == var.name:
                    return var, False

    function_parser: FunctionSolc | None = (
        caller_context if isinstance(caller_context, FunctionSolc) else None
    )
    # Use ret0/ret1 to help mypy
    ret0 = _find_variable_from_ref_declaration(
        referenced_declaration,
        direct_contracts,
        direct_functions,
        function_parser,
        contract_declarer,
    )
    if ret0:
        return ret0, False

    ret1 = _find_variable_in_function_parser(var_name, function_parser)
    if ret1:
        return ret1, False

    ret = _find_in_contract(var_name, contract, contract_declarer, is_super, is_identifier_path)
    if ret:
        return ret, False

    # Could refer to any enum
    all_enumss = [c.enums_as_dict for c in current_scope.contracts.values()]
    all_enums = {k: v for d in all_enumss for k, v in d.items()}
    if var_name in all_enums:
        return all_enums[var_name], False

    contracts = current_scope.contracts
    if var_name in contracts:
        return contracts[var_name], False

    if var_name in SOLIDITY_VARIABLES:
        return SolidityVariable(var_name), False

    if var_name in SOLIDITY_FUNCTIONS:
        return SolidityFunction(var_name), False

    # Top level must be at the end, if nothing else was found
    ret, var_was_created = find_top_level(var_name, current_scope)
    if ret:
        return ret, var_was_created

    # Look from reference declaration in all the contracts at the end
    # Because they are many instances where this can't be trusted
    # For example in
    # contract A{
    #     function _f() internal view returns(uint){
    #         return 1;
    #     }
    #
    #     function get() public view returns(uint){
    #         return _f();
    #     }
    # }
    #
    # contract B is A{
    #     function _f() internal view returns(uint){
    #         return 2;
    #     }
    #
    # }
    # get's AST will say that the ref declaration for _f() is A._f(), but in the context of B, its not

    ret = _find_variable_from_ref_declaration(
        referenced_declaration,
        list(current_scope.contracts.values()),
        list(current_scope.functions),
        None,
        None,
    )
    if ret:
        return ret, False

    raise VariableNotFound(
        f"Variable not found: {var_name} (context {contract} {contract.source_mapping.to_detailed_str()})"
    )



from slither.tools.mutator.mutators.MVIV import MVIV  # severity low
from slither.tools.mutator.mutators.MVIE import MVIE  # severity low
from slither.tools.mutator.mutators.LOR import LOR  # severity medium
from slither.tools.mutator.mutators.UOR import UOR  # severity medium
from slither.tools.mutator.mutators.SBR import SBR  # severity medium
from slither.tools.mutator.mutators.AOR import AOR  # severity medium
from slither.tools.mutator.mutators.BOR import BOR  # severity medium
from slither.tools.mutator.mutators.ASOR import ASOR  # severity medium
from slither.tools.mutator.mutators.MWA import MWA  # severity medium
from slither.tools.mutator.mutators.LIR import LIR  # severity medium
from slither.tools.mutator.mutators.FHR import FHR  # severity medium
from slither.tools.mutator.mutators.MIA import MIA  # severity medium
from slither.tools.mutator.mutators.ROR import ROR  # severity medium
from slither.tools.mutator.mutators.RR import RR  # severity high
from slither.tools.mutator.mutators.CR import CR  # severity high



"""
Module detecting public mappings with nested variables (returns incorrect values prior to 0.5.x)
"""

from typing import Any
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    ALL_SOLC_VERSIONS_04,
)
from slither.core.solidity_types.mapping_type import MappingType
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.core.declarations.structure import Structure
from slither.core.declarations.contract import Contract
from slither.core.variables.state_variable import StateVariable
from slither.utils.output import Output


def detect_public_nested_mappings(contract: Contract) -> list[StateVariable | Any]:
    """
    Detect any state variables that are initialized from an immediate function call (prior to constructor run).
    :param contract: The contract to detect state variable definitions for.
    :return: A list of all state variables defined in the given contract that meet the specified criteria.
    """
    results = []

    for state_variable in contract.variables:
        # Verify this variable is defined in this contract
        if state_variable.contract != contract:
            continue

        # Verify this is a public mapping
        if state_variable.visibility != "public" or not isinstance(
            state_variable.type, MappingType
        ):
            continue

        # Verify the value type is a user defined type (struct)
        if not isinstance(state_variable.type.type_to, UserDefinedType) or not isinstance(
            state_variable.type.type_to.type, Structure
        ):
            continue

        # Obtain the struct
        struct_type = state_variable.type.type_to.type
        for struct_member in struct_type.elems.values():
            if isinstance(struct_member.type, UserDefinedType) and isinstance(
                struct_member.type.type, Structure
            ):
                results.append(state_variable)
                break

    return results


class PublicMappingNested(AbstractDetector):
    """
    Detects public mappings with nested variables (returns incorrect values prior to 0.5.x)
    """

    ARGUMENT = "public-mappings-nested"
    HELP = "Public mappings with nested variables"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#public-mappings-with-nested-variables"

    WIKI_TITLE = "Public mappings with nested variables"
    WIKI_DESCRIPTION = "Prior to Solidity 0.5, a public mapping with nested structures returned [incorrect values](https://github.com/ethereum/solidity/issues/5520)."
    WIKI_EXPLOIT_SCENARIO = """Bob interacts with a contract that has a public mapping with nested structures. The values returned by the mapping are incorrect, breaking Bob's usage"""
    WIKI_RECOMMENDATION = "Do not use public mapping with nested structures."

    VULNERABLE_SOLC_VERSIONS = ALL_SOLC_VERSIONS_04

    def _detect(self) -> list[Output]:
        """
        Detect public mappings with nested variables (returns incorrect values prior to 0.5.x)

        Returns:
            list: {'vuln', 'filename,'contract','func', 'public_nested_mappings'}

        """
        results = []

        for contract in self.contracts:
            public_nested_mappings = detect_public_nested_mappings(contract)
            if public_nested_mappings:
                for public_nested_mapping in public_nested_mappings:
                    info = [public_nested_mapping, " is a public mapping with nested variables\n"]
                    json = self.generate_result(info)
                    results.append(json)

        return results



import re
import logging

from collections.abc import Callable, Sequence

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.variables import Variable
from slither.slithir.operations import (
    Send,
    Transfer,
    OperationWithLValue,
    HighLevelCall,
    LowLevelCall,
    InternalCall,
    InternalDynamicCall,
    Operation,
)
from slither.core.declarations import Modifier, Event
from slither.core.solidity_types import UserDefinedType, MappingType
from slither.core.declarations import Enum, Contract, Structure, Function
from slither.core.solidity_types.elementary_type import ElementaryTypeName
from slither.core.variables.local_variable import LocalVariable
from slither.formatters.exceptions import FormatError, FormatImpossible
from slither.formatters.utils.patches import create_patch

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("Slither.Format")


def custom_format(compilation_unit: SlitherCompilationUnit, result: dict) -> None:
    elements = result["elements"]
    for element in elements:
        target = element["additional_fields"]["target"]

        convention = element["additional_fields"]["convention"]

        if convention == "l_O_I_should_not_be_used":
            # l_O_I_should_not_be_used cannot be automatically patched
            logger.info(
                f"The following naming convention cannot be patched: \n{result['description']}"
            )
            continue

        _patch(compilation_unit, result, element, target)


# endregion
###################################################################################
###################################################################################
# region Conventions
###################################################################################
###################################################################################

KEY = "ALL_NAMES_USED"

# https://solidity.readthedocs.io/en/v0.5.11/miscellaneous.html#reserved-keywords
SOLIDITY_KEYWORDS = [
    "abstract",
    "after",
    "alias",
    "apply",
    "auto",
    "case",
    "catch",
    "copyof",
    "default",
    "define",
    "final",
    "immutable",
    "implements",
    "in",
    "inline",
    "let",
    "macro",
    "match",
    "mutable",
    "null",
    "of",
    "override",
    "partial",
    "promise",
    "reference",
    "relocatable",
    "sealed",
    "sizeof",
    "static",
    "supports",
    "switch",
    "try",
    "typedef",
    "typeof",
    "unchecked",
]

# https://solidity.readthedocs.io/en/v0.5.11/miscellaneous.html#language-grammar
SOLIDITY_KEYWORDS += [
    "pragma",
    "import",
    "contract",
    "library",
    "contract",
    "function",
    "using",
    "struct",
    "enum",
    "public",
    "private",
    "internal",
    "external",
    "calldata",
    "memory",
    "modifier",
    "view",
    "pure",
    "constant",
    "storage",
    "for",
    "if",
    "while",
    "break",
    "return",
    "throw",
    "else",
    "type",
]

SOLIDITY_KEYWORDS += ElementaryTypeName


def _name_already_use(slither: SlitherCompilationUnit, name: str) -> bool:
    # Do not convert to a name used somewhere else
    if KEY not in slither.context:
        all_names: set[str] = set()
        for contract in slither.contracts_derived:
            all_names = all_names.union({st.name for st in contract.structures})
            all_names = all_names.union({f.name for f in contract.functions_and_modifiers})
            all_names = all_names.union({e.name for e in contract.enums})
            all_names = all_names.union({s.name for s in contract.state_variables if s.name})

            for function in contract.functions:
                all_names = all_names.union({v.name for v in function.variables if v.name})

        slither.context[KEY] = all_names
    return name in slither.context[KEY]


def _convert_CapWords(original_name: str, slither: SlitherCompilationUnit) -> str:
    name = original_name.capitalize()

    while "_" in name:
        offset = name.find("_")
        if len(name) > offset:
            name = name[0:offset] + name[offset + 1].upper() + name[offset + 1 :]

    if _name_already_use(slither, name):
        raise FormatImpossible(f"{original_name} cannot be converted to {name} (already used)")

    if name in SOLIDITY_KEYWORDS:
        raise FormatImpossible(f"{original_name} cannot be converted to {name} (Solidity keyword)")
    return name


def _convert_mixedCase(original_name: str | bytes, compilation_unit: SlitherCompilationUnit) -> str:
    if isinstance(original_name, bytes):
        name = original_name.decode("utf8")
    else:
        name = original_name

    while "_" in name:
        offset = name.find("_")
        if len(name) > offset:
            name = name[0:offset] + name[offset + 1].upper() + name[offset + 2 :]

    name = name[0].lower() + name[1:]
    if _name_already_use(compilation_unit, name):
        raise FormatImpossible(f"{original_name} cannot be converted to {name} (already used)")  # type: ignore
    if name in SOLIDITY_KEYWORDS:
        raise FormatImpossible(f"{original_name} cannot be converted to {name} (Solidity keyword)")  # type: ignore
    return name


def _convert_UPPER_CASE_WITH_UNDERSCORES(
    name: str, compilation_unit: SlitherCompilationUnit
) -> str:
    if _name_already_use(compilation_unit, name.upper()):
        raise FormatImpossible(f"{name} cannot be converted to {name.upper()} (already used)")
    if name.upper() in SOLIDITY_KEYWORDS:
        raise FormatImpossible(f"{name} cannot be converted to {name.upper()} (Solidity keyword)")
    return name.upper()


TARGET_TYPE = Contract | Variable | Function
CONVENTION_F_TYPE = Callable[[str, SlitherCompilationUnit], str]

conventions: dict[str, CONVENTION_F_TYPE] = {
    "CapWords": _convert_CapWords,
    "mixedCase": _convert_mixedCase,
    "UPPER_CASE_WITH_UNDERSCORES": _convert_UPPER_CASE_WITH_UNDERSCORES,
}


# endregion
###################################################################################
###################################################################################
# region Helpers
###################################################################################
###################################################################################


def _get_from_contract(
    compilation_unit: SlitherCompilationUnit, element: dict, name: str, getter: str
) -> TARGET_TYPE:
    scope = compilation_unit.get_scope(element["source_mapping"]["filename_absolute"])
    contract_name = element["type_specific_fields"]["parent"]["name"]
    contract = scope.get_contract_from_name(contract_name)
    return getattr(contract, getter)(name)


# endregion
###################################################################################
###################################################################################
# region Patch dispatcher
###################################################################################
###################################################################################


def _patch(
    compilation_unit: SlitherCompilationUnit, result: dict, element: dict, _target: str
) -> None:
    scope = compilation_unit.get_scope(element["source_mapping"]["filename_absolute"])

    target: TARGET_TYPE | None = None

    if _target == "contract":
        target = scope.get_contract_from_name(element["name"])

    elif _target == "structure":
        target = _get_from_contract(
            compilation_unit, element, element["name"], "get_structure_from_name"
        )

    elif _target == "event":
        target = _get_from_contract(
            compilation_unit, element, element["name"], "get_event_from_name"
        )

    elif _target == "function":
        # Avoid constructor (FP?)
        if element["name"] != element["type_specific_fields"]["parent"]["name"]:
            function_sig = element["type_specific_fields"]["signature"]
            target = _get_from_contract(
                compilation_unit, element, function_sig, "get_function_from_signature"
            )

    elif _target == "modifier":
        modifier_sig = element["type_specific_fields"]["signature"]
        target = _get_from_contract(
            compilation_unit, element, modifier_sig, "get_modifier_from_signature"
        )

    elif _target == "parameter":
        contract_name = element["type_specific_fields"]["parent"]["type_specific_fields"]["parent"][
            "name"
        ]
        function_sig = element["type_specific_fields"]["parent"]["type_specific_fields"][
            "signature"
        ]
        param_name = element["name"]
        contract = scope.get_contract_from_name(contract_name)
        assert contract
        function = contract.get_function_from_full_name(function_sig)
        assert function
        target = function.get_local_variable_from_name(param_name)

    elif _target in ["variable", "variable_constant"]:
        # Local variable
        if element["type_specific_fields"]["parent"] == "function":
            contract_name = element["type_specific_fields"]["parent"]["type_specific_fields"][
                "parent"
            ]["name"]
            function_sig = element["type_specific_fields"]["parent"]["type_specific_fields"][
                "signature"
            ]
            var_name = element["name"]
            contract = scope.get_contract_from_name(contract_name)
            assert contract
            function = contract.get_function_from_full_name(function_sig)
            assert function
            target = function.get_local_variable_from_name(var_name)
        # State variable
        else:
            target = _get_from_contract(
                compilation_unit, element, element["name"], "get_state_variable_from_name"
            )

    elif _target == "enum":
        target = _get_from_contract(
            compilation_unit, element, element["name"], "get_enum_from_canonical_name"
        )

    else:
        raise FormatError("Unknown naming convention! " + _target)

    assert target
    _explore(
        compilation_unit, result, target, conventions[element["additional_fields"]["convention"]]
    )


# endregion
###################################################################################
###################################################################################
# region Explore functions
###################################################################################
###################################################################################

# group 1: beginning of the from type
# group 2: beginning of the to type
# nested mapping are within the group 1
# RE_MAPPING = '[ ]*mapping[ ]*\([ ]*([\=\>\(\) a-zA-Z0-9\._\[\]]*)[ ]*=>[ ]*([a-zA-Z0-9\._\[\]]*)\)'
RE_MAPPING_FROM = rb"([a-zA-Z0-9\._\[\]]*)"
RE_MAPPING_TO = rb"([\=\>\(\) a-zA-Z0-9\._\[\]\   ]*)"
RE_MAPPING = (
    rb"[ ]*mapping[ ]*\([ ]*" + RE_MAPPING_FROM + b"[ ]*" + b"=>" + b"[ ]*" + RE_MAPPING_TO + rb"\)"
)


def _is_var_declaration(slither: SlitherCompilationUnit, filename: str, start: int) -> bool:
    """
    Detect usage of 'var ' for Solidity < 0.5
    :param slither:
    :param filename:
    :param start:
    :return:
    """
    v = "var "
    return (
        slither.core.source_code[filename].encode("utf8")[start : start + len(v)].decode("utf8")
        == v
    )


def _explore_type(
    slither: SlitherCompilationUnit,
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
    custom_type: type | list[type] | None,
    filename_source_code: str,
    start: int,
    end: int,
) -> None:
    if isinstance(custom_type, UserDefinedType):
        # Patch type based on contract/enum
        if isinstance(custom_type.type, (Enum, Contract)):
            if custom_type.type == target:
                old_str = custom_type.type.name
                new_str = convert(old_str, slither)

                loc_start = start
                if _is_var_declaration(slither, filename_source_code, start):
                    loc_end = loc_start + len("var")
                else:
                    loc_end = loc_start + len(old_str)

                create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)

        else:
            # Patch type based on structure
            assert isinstance(custom_type.type, Structure)
            if custom_type.type == target:
                old_str = custom_type.type.name
                new_str = convert(old_str, slither)

                loc_start = start
                if _is_var_declaration(slither, filename_source_code, start):
                    loc_end = loc_start + len("var")
                else:
                    loc_end = loc_start + len(old_str)

                create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)

            # Structure contain a list of elements, that might need patching
            # .elems return a list of VariableStructure
            _explore_variables_declaration(
                slither, list(custom_type.type.elems.values()), result, target, convert
            )

    if isinstance(custom_type, MappingType):
        # Mapping has three steps:
        # Convert the "from" type
        # Convert the "to" type
        # Convert nested type in the "to"
        # Ex: mapping (mapping (badName => uint) => uint)

        # Do the comparison twice, so we can factor together the re matching
        # mapping can only have elementary type in type_from
        if isinstance(custom_type.type_to, (UserDefinedType, MappingType)) or target in [
            custom_type.type_from,
            custom_type.type_to,
        ]:
            full_txt_start = start
            full_txt_end = end
            full_txt = slither.core.source_code[filename_source_code].encode("utf8")[
                full_txt_start:full_txt_end
            ]
            re_match = re.match(RE_MAPPING, full_txt)
            assert re_match

            if custom_type.type_from == target:
                old_str = custom_type.type_from.name
                new_str = convert(old_str, slither)

                loc_start = start + re_match.start(1)
                loc_end = loc_start + len(old_str)

                create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)

            if custom_type.type_to == target:
                old_str = custom_type.type_to.name
                new_str = convert(old_str, slither)

                loc_start = start + re_match.start(2)
                loc_end = loc_start + len(old_str)

                create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)

            if isinstance(custom_type.type_to, (UserDefinedType, MappingType)):
                loc_start = start + re_match.start(2)
                loc_end = start + re_match.end(2)
                _explore_type(
                    slither,
                    result,
                    target,
                    convert,
                    custom_type.type_to,
                    filename_source_code,
                    loc_start,
                    loc_end,
                )


def _explore_variables_declaration(
    slither: SlitherCompilationUnit,
    variables: Sequence[Variable],
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
    patch_comment: bool = False,
) -> None:
    for variable in variables:
        # First explore the type of the variable
        filename_source_code = variable.source_mapping.filename.absolute
        full_txt_start = variable.source_mapping.start
        full_txt_end = full_txt_start + variable.source_mapping.length
        full_txt = slither.core.source_code[filename_source_code].encode("utf8")[
            full_txt_start:full_txt_end
        ]

        _explore_type(
            slither,
            result,
            target,
            convert,
            variable.type,
            filename_source_code,
            full_txt_start,
            variable.source_mapping.start + variable.source_mapping.length,
        )

        # If the variable is the target
        if variable == target:
            old_str = variable.name
            if old_str is None:
                old_str = ""
            new_str = convert(old_str, slither)

            loc_start = full_txt_start + full_txt.find(old_str.encode("utf8"))
            loc_end = loc_start + len(old_str)

            create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)

            # Patch comment only makes sense for local variable declaration in the parameter list
            if patch_comment and isinstance(variable, LocalVariable):
                if variable.source_mapping.lines:
                    func = variable.function
                    end_line = func.source_mapping.lines[0]
                    if variable in func.parameters:
                        idx = len(func.parameters) - func.parameters.index(variable) + 1
                        first_line = end_line - idx - 2

                        potential_comments_ = slither.core.source_code[filename_source_code].encode(
                            "utf8"
                        )
                        potential_comments = potential_comments_.splitlines(keepends=True)[
                            first_line : end_line - 1
                        ]

                        idx_beginning = func.source_mapping.start
                        idx_beginning += -func.source_mapping.starting_column + 1
                        idx_beginning += -sum(len(c) for c in potential_comments)

                        old_comment = f"@param {old_str}".encode()

                        for line in potential_comments:
                            idx = line.find(old_comment)
                            if idx >= 0:
                                loc_start = idx + idx_beginning
                                loc_end = loc_start + len(old_comment)
                                new_comment = f"@param {new_str}".encode()

                                create_patch(
                                    result,
                                    filename_source_code,
                                    loc_start,
                                    loc_end,
                                    old_comment,
                                    new_comment,
                                )

                                break
                            idx_beginning += len(line)


def _explore_structures_declaration(
    slither: SlitherCompilationUnit,
    structures: Sequence[Structure],
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
) -> None:
    for st in structures:
        # Explore the variable declared within the structure (VariableStructure)
        _explore_variables_declaration(slither, list(st.elems.values()), result, target, convert)

        # If the structure is the target
        if st == target:
            old_str = st.name
            new_str = convert(old_str, slither)

            filename_source_code = st.source_mapping.filename.absolute
            full_txt_start = st.source_mapping.start
            full_txt_end = full_txt_start + st.source_mapping.length
            full_txt = slither.core.source_code[filename_source_code].encode("utf8")[
                full_txt_start:full_txt_end
            ]

            # The name is after the space
            matches = re.finditer(b"struct[ ]*", full_txt)
            # Look for the end offset of the largest list of ' '
            loc_start = full_txt_start + max(matches, key=lambda x: len(x.group())).end()
            loc_end = loc_start + len(old_str)

            create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)


def _explore_events_declaration(
    slither: SlitherCompilationUnit,
    events: Sequence[Event],
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
) -> None:
    for event in events:
        # Explore the parameters
        _explore_variables_declaration(slither, event.elems, result, target, convert)

        # If the event is the target
        if event == target:
            filename_source_code = event.source_mapping.filename.absolute

            old_str = event.name
            new_str = convert(old_str, slither)

            loc_start = event.source_mapping.start
            loc_end = loc_start + len(old_str)

            create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)


def get_ir_variables(ir: Operation) -> list[Variable | Function]:
    all_vars = ir.read

    if isinstance(ir, (InternalCall, InternalDynamicCall, HighLevelCall)):
        all_vars += [ir.function]

    if isinstance(ir, (HighLevelCall, Send, LowLevelCall, Transfer)):
        all_vars += [ir.call_value]

    if isinstance(ir, (HighLevelCall, LowLevelCall)):
        all_vars += [ir.call_gas]

    if isinstance(ir, OperationWithLValue):
        all_vars += [ir.lvalue]

    return [v for v in all_vars if v]


def _explore_irs(
    slither: SlitherCompilationUnit,
    irs: list[Operation],
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
) -> None:
    if not irs:
        return
    for ir in irs:
        for v in get_ir_variables(ir):
            if target == v or (
                isinstance(target, Function)
                and isinstance(v, Function)
                and v.canonical_name == target.canonical_name
            ):
                source_mapping = ir.expression.source_mapping
                filename_source_code = source_mapping.filename.absolute
                full_txt_start = source_mapping.start
                full_txt_end = full_txt_start + source_mapping.length
                full_txt = slither.core.source_code[filename_source_code].encode("utf8")[
                    full_txt_start:full_txt_end
                ]

                if target.name.encode("utf8") not in full_txt:
                    raise FormatError(f"{target} not found in {full_txt} ({source_mapping}")

                old_str = target.name.encode("utf8")
                new_str = convert(old_str, slither)

                counter = 0
                # Can be found multiple time on the same IR
                # We patch one by one
                while old_str in full_txt:
                    target_found_at = full_txt.find(old_str)

                    full_txt = full_txt[target_found_at + 1 :]
                    counter += target_found_at

                    loc_start = full_txt_start + counter
                    loc_end = loc_start + len(old_str)

                    create_patch(
                        result,
                        filename_source_code,
                        loc_start,
                        loc_end,
                        old_str,
                        new_str,
                    )


def _explore_functions(
    slither: SlitherCompilationUnit,
    functions: list[Function],
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
) -> None:
    for function in functions:
        _explore_variables_declaration(slither, function.variables, result, target, convert, True)
        _explore_irs(slither, function.all_slithir_operations(), result, target, convert)

        if isinstance(target, Function) and function.canonical_name == target.canonical_name:
            old_str = function.name
            new_str = convert(old_str, slither)

            filename_source_code = function.source_mapping.filename.absolute
            full_txt_start = function.source_mapping.start
            full_txt_end = full_txt_start + function.source_mapping.length
            full_txt = slither.core.source_code[filename_source_code].encode("utf8")[
                full_txt_start:full_txt_end
            ]

            # The name is after the space
            if isinstance(target, Modifier):
                matches = re.finditer(b"modifier([ ]*)", full_txt)
            else:
                matches = re.finditer(b"function([ ]*)", full_txt)
            # Look for the end offset of the largest list of ' '
            loc_start = full_txt_start + max(matches, key=lambda x: len(x.group())).end()
            loc_end = loc_start + len(old_str)

            create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)


def _explore_enums(
    slither: SlitherCompilationUnit,
    enums: Sequence[Enum],
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
) -> None:
    for enum in enums:
        if enum == target:
            old_str = enum.name
            new_str = convert(old_str, slither)

            filename_source_code = enum.source_mapping.filename.absolute
            full_txt_start = enum.source_mapping.start
            full_txt_end = full_txt_start + enum.source_mapping.length
            full_txt = slither.core.source_code[filename_source_code].encode("utf8")[
                full_txt_start:full_txt_end
            ]

            # The name is after the space
            matches = re.finditer(b"enum([ ]*)", full_txt)
            # Look for the end offset of the largest list of ' '
            loc_start = full_txt_start + max(matches, key=lambda x: len(x.group())).end()
            loc_end = loc_start + len(old_str)

            create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)


def _explore_contract(
    slither: SlitherCompilationUnit,
    contract: Contract,
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
) -> None:
    _explore_variables_declaration(slither, contract.state_variables, result, target, convert)
    _explore_structures_declaration(slither, contract.structures, result, target, convert)
    _explore_functions(slither, contract.functions_and_modifiers, result, target, convert)
    _explore_enums(slither, contract.enums, result, target, convert)

    if contract == target:
        filename_source_code = contract.source_mapping.filename.absolute
        full_txt_start = contract.source_mapping.start
        full_txt_end = full_txt_start + contract.source_mapping.length
        full_txt = slither.core.source_code[filename_source_code].encode("utf8")[
            full_txt_start:full_txt_end
        ]

        old_str = contract.name
        new_str = convert(old_str, slither)

        # The name is after the space
        matches = re.finditer(b"contract[ ]*", full_txt)
        # Look for the end offset of the largest list of ' '
        loc_start = full_txt_start + max(matches, key=lambda x: len(x.group())).end()

        loc_end = loc_start + len(old_str)

        create_patch(result, filename_source_code, loc_start, loc_end, old_str, new_str)


def _explore(
    compilation_unit: SlitherCompilationUnit,
    result: dict,
    target: TARGET_TYPE,
    convert: CONVENTION_F_TYPE,
) -> None:
    for contract in compilation_unit.contracts_derived:
        _explore_contract(compilation_unit, contract, result, target, convert)


# endregion



import argparse
from contextlib import contextmanager
import os
from pathlib import Path
import shutil
import subprocess
import tempfile
from unittest import mock

import pytest
from slither import Slither
from slither.tools.mutator.__main__ import _get_mutators, main, parse_target_selectors
from slither.tools.mutator.utils.testing_generated_mutant import run_test_cmd
from slither.tools.mutator.utils.file_handling import get_sol_file_list, backup_source_file
from slither.utils.function import get_function_id
from slither.tools.mutator.mutators.RR import RR


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"

foundry_available = shutil.which("forge") is not None
project_ready = Path(TEST_DATA_DIR, "test_source_unit/lib/forge-std").exists()


@contextmanager
def change_directory(new_dir):
    original_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(original_dir)


def test_get_mutators():
    mutators = _get_mutators(None)
    assert mutators

    mutators = _get_mutators(["ASOR"])
    assert len(mutators) == 1
    assert mutators[0].NAME == "ASOR"

    mutators = _get_mutators(["ASOR", "NotExisiting"])
    assert len(mutators) == 1


@mock.patch(
    "argparse.ArgumentParser.parse_args",
    return_value=argparse.Namespace(
        test_cmd="forge test",
        test_dir=None,
        ignore_dirs="lib,mutation_campaign",
        output_dir=None,
        timeout=None,
        solc_remaps="forge-std=./lib/forge-std",
        verbose=None,
        very_verbose=None,
        mutators_to_run=None,
        comprehensive=None,
        codebase=(TEST_DATA_DIR / "test_source_unit" / "src" / "Counter.sol").as_posix(),
        contract_names="Counter",
    ),
)
@pytest.mark.skip(reason="Slow test")
def test_mutator(mock_args, solc_binary_path):
    with change_directory(TEST_DATA_DIR / "test_source_unit"):
        main()


def test_backup_source_file(solc_binary_path):
    solc_path = solc_binary_path("0.8.15")

    file_path = (TEST_DATA_DIR / "test_source_unit" / "src" / "Counter.sol").as_posix()
    sl = Slither(file_path, solc=solc_path, compile_force_framework="solc")

    with tempfile.TemporaryDirectory() as directory:
        files_dict = backup_source_file(sl.source_code, Path(directory))

        assert len(files_dict) == 1
        assert Path(files_dict[file_path]).exists()


@pytest.mark.skipif(
    not foundry_available or not project_ready, reason="requires Foundry and project setup"
)
def test_get_sol_file_list():
    project_directory = TEST_DATA_DIR / "test_source_unit"

    files = get_sol_file_list(project_directory, None)

    assert len(files) == 46

    files = get_sol_file_list(project_directory, ["lib"])
    assert len(files) == 3

    files = get_sol_file_list(project_directory, ["lib", "script"])
    assert len(files) == 2

    files = get_sol_file_list(project_directory / "src" / "Counter.sol", None)
    assert len(files) == 1

    (project_directory / "test.sol").mkdir()
    files = get_sol_file_list(project_directory, None)
    assert all("test.sol" not in file for file in files)
    (project_directory / "test.sol").rmdir()


@pytest.mark.skipif(
    not foundry_available or not project_ready, reason="requires Foundry and project setup"
)
def test_run_test(caplog):
    with change_directory(TEST_DATA_DIR / "test_source_unit"):
        result = run_test_cmd("forge test", timeout=None, target_file=None, verbose=True)
        assert result
        assert not caplog.records

        # Failed command
        result = run_test_cmd("forge non-test", timeout=None, target_file=None, verbose=True)
        assert not result
        assert caplog.records


def test_run_tests_timeout(caplog, monkeypatch):
    def mock_run(*args, **kwargs):
        raise subprocess.TimeoutExpired(cmd=args[0], timeout=kwargs.get("timeout"))

    monkeypatch.setattr(subprocess, "run", mock_run)

    with change_directory(TEST_DATA_DIR / "test_source_unit"):
        result = run_test_cmd("forge test", timeout=1)
        assert not result
        assert "Tests took too long" in caplog.messages[0]


def test_parse_target_selectors_hex():
    """Test hex selector parsing"""
    selectors = parse_target_selectors("0xa9059cbb")
    assert 0xA9059CBB in selectors
    assert len(selectors) == 1


def test_parse_target_selectors_signature():
    """Test signature to selector conversion"""
    selectors = parse_target_selectors("transfer(address,uint256)")
    expected = get_function_id("transfer(address,uint256)")
    assert expected in selectors
    assert len(selectors) == 1


def test_parse_target_selectors_multiple():
    """Test multiple selectors (mixed formats)"""
    selectors = parse_target_selectors("0xa9059cbb,mint(address,uint256)")
    assert len(selectors) == 2
    assert 0xA9059CBB in selectors
    assert get_function_id("mint(address,uint256)") in selectors


def test_parse_target_selectors_empty_parts():
    """Test handling of empty parts in comma-separated list"""
    selectors = parse_target_selectors("0xa9059cbb,,0x40c10f19")
    assert len(selectors) == 2


def test_parse_target_selectors_whitespace():
    """Test handling of whitespace around selectors"""
    # Note: 0xa9059cbb IS transfer(address,uint256), so deduped to 1
    selectors = parse_target_selectors("  0xa9059cbb  ,  mint(address,uint256)  ")
    assert len(selectors) == 2


def test_should_mutate_function_no_filter(solc_binary_path):
    """When target_selectors is None, all functions should be mutated"""
    solc_path = solc_binary_path("0.8.15")
    file_path = (TEST_DATA_DIR / "test_source_unit" / "src" / "Counter.sol").as_posix()
    sl = Slither(file_path, solc=solc_path, compile_force_framework="solc")

    contract = next(c for c in sl.contracts if c.name == "Counter")

    with tempfile.TemporaryDirectory() as tmpdir:
        mutator = RR(
            sl.compilation_units[0],
            timeout=30,
            testing_command="true",
            testing_directory=None,
            contract_instance=contract,
            solc_remappings=None,
            verbose=False,
            output_folder=Path(tmpdir),
            dont_mutate_line=[],
            target_selectors=None,
            target_modifiers=None,
        )

        # All functions should be included when no filter
        for func in contract.functions_declared:
            assert mutator.should_mutate_function(func) is True


def test_should_mutate_function_matching_selector(solc_binary_path):
    """Function matching target selector should be mutated"""
    solc_path = solc_binary_path("0.8.15")
    file_path = (TEST_DATA_DIR / "test_source_unit" / "src" / "Counter.sol").as_posix()
    sl = Slither(file_path, solc=solc_path, compile_force_framework="solc")

    contract = next(c for c in sl.contracts if c.name == "Counter")
    increment_selector = get_function_id("increment()")

    with tempfile.TemporaryDirectory() as tmpdir:
        mutator = RR(
            sl.compilation_units[0],
            timeout=30,
            testing_command="true",
            testing_directory=None,
            contract_instance=contract,
            solc_remappings=None,
            verbose=False,
            output_folder=Path(tmpdir),
            dont_mutate_line=[],
            target_selectors={increment_selector},
            target_modifiers=None,
        )

        for func in contract.functions_declared:
            if func.name == "increment":
                assert mutator.should_mutate_function(func) is True
            else:
                assert mutator.should_mutate_function(func) is False


def test_should_mutate_function_no_match(solc_binary_path):
    """Function not matching any selector should not be mutated"""
    solc_path = solc_binary_path("0.8.15")
    file_path = (TEST_DATA_DIR / "test_source_unit" / "src" / "Counter.sol").as_posix()
    sl = Slither(file_path, solc=solc_path, compile_force_framework="solc")

    contract = next(c for c in sl.contracts if c.name == "Counter")

    with tempfile.TemporaryDirectory() as tmpdir:
        mutator = RR(
            sl.compilation_units[0],
            timeout=30,
            testing_command="true",
            testing_directory=None,
            contract_instance=contract,
            solc_remappings=None,
            verbose=False,
            output_folder=Path(tmpdir),
            dont_mutate_line=[],
            target_selectors={0xDEADBEEF},
            target_modifiers=None,
        )

        # No functions should match bogus selector
        for func in contract.functions_declared:
            assert mutator.should_mutate_function(func) is False


def test_should_mutate_function_includes_modifier(solc_binary_path):
    """Modifier used by target function should be mutated"""
    solc_path = solc_binary_path("0.8.15")
    file_path = (TEST_DATA_DIR / "test_source_unit" / "src" / "Counter.sol").as_posix()
    sl = Slither(file_path, solc=solc_path, compile_force_framework="solc")

    contract = next(c for c in sl.contracts if c.name == "Counter")
    restricted_selector = get_function_id("restrictedIncrement()")

    with tempfile.TemporaryDirectory() as tmpdir:
        mutator = RR(
            sl.compilation_units[0],
            timeout=30,
            testing_command="true",
            testing_directory=None,
            contract_instance=contract,
            solc_remappings=None,
            verbose=False,
            output_folder=Path(tmpdir),
            dont_mutate_line=[],
            target_selectors={restricted_selector},
            target_modifiers={"onlyOwner"},
        )

        # onlyOwner modifier should be included
        for mod in contract.modifiers:
            if mod.name == "onlyOwner":
                assert mutator.should_mutate_function(mod) is True



import logging
from slither.core.declarations import Contract, Function
from slither.core.cfg.node import Node
from slither.utils.function import get_function_id
from slither.exceptions import SlitherError
from slither.analyses.evm.evm_cfg_builder import load_evm_cfg_builder

logger = logging.getLogger("ConvertToEVM")

KEY_EVM_INS = "EVM_INSTRUCTIONS"


def get_evm_instructions(obj):
    assert isinstance(obj, (Function, Contract, Node))

    if KEY_EVM_INS not in obj.context:
        CFG = load_evm_cfg_builder()

        slither = obj.slither

        if not slither.crytic_compile:
            raise SlitherError("EVM features require to compile with crytic-compile")

        contract_info = {}
        function_info = {}
        node_info = {}

        if isinstance(obj, Node):
            contract_info["contract"] = obj.function.contract
        elif isinstance(obj, Function):
            contract_info["contract"] = obj.contract
        else:
            contract_info["contract"] = obj

        # Get contract runtime bytecode, srcmap and cfg
        contract_info["bytecode_runtime"] = slither.crytic_compile.bytecode_runtime(
            contract_info["contract"].name
        )
        contract_info["srcmap_runtime"] = slither.crytic_compile.srcmap_runtime(
            contract_info["contract"].name
        )
        contract_info["cfg"] = CFG(contract_info["bytecode_runtime"])

        # Get contract init bytecode, srcmap and cfg
        contract_info["bytecode_init"] = slither.crytic_compile.bytecode_init(
            contract_info["contract"].name
        )
        contract_info["srcmap_init"] = slither.crytic_compile.srcmap_init(
            contract_info["contract"].name
        )
        contract_info["cfg_init"] = CFG(contract_info["bytecode_init"])

        # Get evm instructions
        if isinstance(obj, Contract):
            # Get evm instructions for contract
            obj.context[KEY_EVM_INS] = _get_evm_instructions_contract(contract_info)

        elif isinstance(obj, Function):
            # Get evm instructions for function
            function_info["function"] = obj
            function_info["contract_info"] = contract_info
            obj.context[KEY_EVM_INS] = _get_evm_instructions_function(function_info)

        else:
            # Get evm instructions for node
            node_info["node"] = obj

            # CFG and srcmap depend on function being constructor or not
            if node_info["node"].function.is_constructor:
                cfg = contract_info["cfg_init"]
                srcmap = contract_info["srcmap_init"]
            else:
                cfg = contract_info["cfg"]
                srcmap = contract_info["srcmap_runtime"]

            node_info["cfg"] = cfg
            node_info["srcmap"] = srcmap
            node_info["contract"] = contract_info["contract"]
            node_info["slither"] = slither

            obj.context[KEY_EVM_INS] = _get_evm_instructions_node(node_info)

    return obj.context.get(KEY_EVM_INS, [])


def _get_evm_instructions_contract(contract_info):
    # Combine the instructions of constructor and the rest of the contract
    return contract_info["cfg_init"].instructions + contract_info["cfg"].instructions


def _get_evm_instructions_function(function_info):
    function = function_info["function"]

    # CFG depends on function being constructor or not
    if function.is_constructor:
        cfg = function_info["contract_info"]["cfg_init"]
        # _dispatcher is the only function recognised by evm-cfg-builder in bytecode_init.
        # _dispatcher serves the role of the constructor in init code,
        #    given that there are no other functions.
        # Todo: Could rename it appropriately in evm-cfg-builder
        #    by detecting that init bytecode is being parsed.
        name = "_dispatcher"
        func_hash = ""
    else:
        cfg = function_info["contract_info"]["cfg"]
        name = function.name
        # Get first four bytes of function singature's keccak-256 hash used as function selector
        func_hash = str(hex(get_function_id(function.full_name)))

    function_evm = _get_function_evm(cfg, name, func_hash)
    if function_evm is None:
        to_log = "Function " + function.name + " not found in the EVM code"
        logger.error(to_log)
        raise SlitherError("Function " + function.name + " not found in the EVM code")

    function_ins = []
    for basic_block in sorted(function_evm.basic_blocks, key=lambda x: x.start.pc):
        for ins in basic_block.instructions:
            function_ins.append(ins)

    return function_ins


def _get_evm_instructions_node(node_info):
    # Get evm instructions for node's contract
    contract_pcs = generate_source_to_evm_ins_mapping(
        node_info["cfg"].instructions,
        node_info["srcmap"],
        node_info["slither"],
        node_info["contract"].source_mapping.filename.absolute,
    )
    contract_file = (
        node_info["slither"]
        .source_code[node_info["contract"].source_mapping.filename.absolute]
        .encode("utf8")
    )

    # Get evm instructions corresponding to node's source line number
    node_source_line = contract_file[0 : node_info["node"].source_mapping.start].count(b"\n") + 1
    node_pcs = contract_pcs.get(node_source_line, [])
    node_ins = []
    for pc in node_pcs:
        node_ins.append(node_info["cfg"].get_instruction_at(pc))

    return node_ins


def _get_function_evm(cfg, function_name, function_hash):
    for function_evm in cfg.functions:
        # Match function hash
        if function_evm.name[:2] == "0x" and function_evm.name == function_hash:
            return function_evm
        # Match function name
        if function_evm.name[:2] != "0x" and function_evm.name.split("(")[0] == function_name:
            return function_evm
    return None


def generate_source_to_evm_ins_mapping(evm_instructions, srcmap_runtime, slither, filename):
    """
    Generate Solidity source to EVM instruction mapping using evm_cfg_builder:cfg.instructions
    and solc:srcmap_runtime

    Returns: Solidity source to EVM instruction mapping
    """

    source_to_evm_mapping = {}
    file_source = slither.source_code[filename].encode("utf8")
    prev_mapping = []

    for idx, mapping in enumerate(srcmap_runtime):
        # Parse srcmap_runtime according to its format
        # See https://solidity.readthedocs.io/en/v0.5.9/miscellaneous.html#source-mappings
        # In order to compress these source mappings especially for bytecode, the following rules are used:
        # If a field is empty, the value of the preceding element is used.
        # If a : is missing, all following fields are considered empty.
        mapping_item = mapping.split(":")
        mapping_item += prev_mapping[len(mapping_item) :]

        for i, _ in enumerate(mapping_item):
            if mapping_item[i] == "":
                mapping_item[i] = prev_mapping[i]

        offset, _, file_id, *_ = mapping_item
        prev_mapping = mapping_item

        if file_id == "-1":
            # Internal compiler-generated code snippets to be ignored
            # See https://github.com/ethereum/solidity/issues/6119#issuecomment-467797635
            continue

        line_number = file_source[0 : int(offset)].count(b"\n") + 1

        # Append evm instructions to the corresponding source line number
        # Note: Some evm instructions in mapping are not necessarily in program execution order
        # Note: The order depends on how solc creates the srcmap_runtime
        source_to_evm_mapping.setdefault(line_number, []).append(evm_instructions[idx].pc)

    return source_to_evm_mapping



import logging
from typing import TYPE_CHECKING, Any

from slither.core import expressions
from slither.core.scope.scope import FileScope
from slither.core.declarations import (
    Function,
    SolidityVariable,
    SolidityVariableComposed,
    SolidityFunction,
    Contract,
    EnumContract,
    EnumTopLevel,
    Enum,
    SolidityImportPlaceHolder,
    Import,
    Structure,
)
from slither.core.expressions import (
    AssignmentOperation,
    AssignmentOperationType,
    UnaryOperationType,
    BinaryOperationType,
    ElementaryTypeNameExpression,
    CallExpression,
    Identifier,
    MemberAccess,
    ConditionalExpression,
    NewElementaryType,
)
from slither.core.expressions.binary_operation import BinaryOperation
from slither.core.expressions.expression import Expression
from slither.core.expressions.index_access import IndexAccess
from slither.core.expressions.literal import Literal
from slither.core.expressions.new_array import NewArray
from slither.core.expressions.new_contract import NewContract
from slither.core.expressions.tuple_expression import TupleExpression
from slither.core.expressions.unary_operation import UnaryOperation
from slither.core.solidity_types import ArrayType, ElementaryType, TypeAlias, UserDefinedType
from slither.core.solidity_types.type import Type
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.local_variable_init_from_tuple import LocalVariableInitFromTuple
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.variable import Variable
from slither.slithir.exceptions import SlithIRError
from slither.slithir.operations import (
    Assignment,
    Binary,
    BinaryType,
    Delete,
    Index,
    InitArray,
    InternalCall,
    Member,
    TypeConversion,
    Unary,
    UnaryType,
    Unpack,
    Return,
    SolidityCall,
    Operation,
)
from slither.slithir.tmp_operations.argument import Argument
from slither.slithir.tmp_operations.tmp_call import TmpCall
from slither.slithir.tmp_operations.tmp_new_array import TmpNewArray
from slither.slithir.tmp_operations.tmp_new_contract import TmpNewContract
from slither.slithir.tmp_operations.tmp_new_elementary_type import TmpNewElementaryType
from slither.slithir.variables import (
    Constant,
    ReferenceVariable,
    TemporaryVariable,
    TupleVariable,
)
from slither.visitors.expression.expression import ExpressionVisitor

if TYPE_CHECKING:
    from slither.core.cfg.node import Node

logger = logging.getLogger("VISTIOR:ExpressionToSlithIR")

key = "expressionToSlithIR"


def get(expression: Expression) -> Any:
    val = expression.context[key]
    # we delete the item to reduce memory use
    del expression.context[key]
    return val


def set_val(expression: Expression, val: Any) -> None:
    expression.context[key] = val


_binary_to_binary = {
    BinaryOperationType.POWER: BinaryType.POWER,
    BinaryOperationType.MULTIPLICATION: BinaryType.MULTIPLICATION,
    BinaryOperationType.DIVISION: BinaryType.DIVISION,
    BinaryOperationType.MODULO: BinaryType.MODULO,
    BinaryOperationType.ADDITION: BinaryType.ADDITION,
    BinaryOperationType.SUBTRACTION: BinaryType.SUBTRACTION,
    BinaryOperationType.LEFT_SHIFT: BinaryType.LEFT_SHIFT,
    BinaryOperationType.RIGHT_SHIFT: BinaryType.RIGHT_SHIFT,
    BinaryOperationType.AND: BinaryType.AND,
    BinaryOperationType.CARET: BinaryType.CARET,
    BinaryOperationType.OR: BinaryType.OR,
    BinaryOperationType.LESS: BinaryType.LESS,
    BinaryOperationType.GREATER: BinaryType.GREATER,
    BinaryOperationType.LESS_EQUAL: BinaryType.LESS_EQUAL,
    BinaryOperationType.GREATER_EQUAL: BinaryType.GREATER_EQUAL,
    BinaryOperationType.EQUAL: BinaryType.EQUAL,
    BinaryOperationType.NOT_EQUAL: BinaryType.NOT_EQUAL,
    BinaryOperationType.ANDAND: BinaryType.ANDAND,
    BinaryOperationType.OROR: BinaryType.OROR,
}


_unary_to_unary = {
    UnaryOperationType.BANG: UnaryType.BANG,
    UnaryOperationType.TILD: UnaryType.TILD,
}


_signed_to_unsigned = {
    BinaryOperationType.DIVISION_SIGNED: BinaryType.DIVISION,
    BinaryOperationType.MODULO_SIGNED: BinaryType.MODULO,
    BinaryOperationType.LESS_SIGNED: BinaryType.LESS,
    BinaryOperationType.GREATER_SIGNED: BinaryType.GREATER,
    BinaryOperationType.RIGHT_SHIFT_ARITHMETIC: BinaryType.RIGHT_SHIFT,
}


def convert_assignment(
    left: LocalVariable | StateVariable | ReferenceVariable,
    right: LocalVariable | StateVariable | ReferenceVariable,
    t: AssignmentOperationType,
    return_type: Type,
) -> Binary | Assignment:
    if t == AssignmentOperationType.ASSIGN:
        return Assignment(left, right, return_type)
    if t == AssignmentOperationType.ASSIGN_OR:
        return Binary(left, left, right, BinaryType.OR)
    if t == AssignmentOperationType.ASSIGN_CARET:
        return Binary(left, left, right, BinaryType.CARET)
    if t == AssignmentOperationType.ASSIGN_AND:
        return Binary(left, left, right, BinaryType.AND)
    if t == AssignmentOperationType.ASSIGN_LEFT_SHIFT:
        return Binary(left, left, right, BinaryType.LEFT_SHIFT)
    if t == AssignmentOperationType.ASSIGN_RIGHT_SHIFT:
        return Binary(left, left, right, BinaryType.RIGHT_SHIFT)
    if t == AssignmentOperationType.ASSIGN_ADDITION:
        return Binary(left, left, right, BinaryType.ADDITION)
    if t == AssignmentOperationType.ASSIGN_SUBTRACTION:
        return Binary(left, left, right, BinaryType.SUBTRACTION)
    if t == AssignmentOperationType.ASSIGN_MULTIPLICATION:
        return Binary(left, left, right, BinaryType.MULTIPLICATION)
    if t == AssignmentOperationType.ASSIGN_DIVISION:
        return Binary(left, left, right, BinaryType.DIVISION)
    if t == AssignmentOperationType.ASSIGN_MODULO:
        return Binary(left, left, right, BinaryType.MODULO)

    raise SlithIRError("Missing type during assignment conversion")


class ExpressionToSlithIR(ExpressionVisitor):
    def __init__(self, expression: Expression, node: "Node") -> None:
        from slither.core.cfg.node import NodeType

        self._expression = expression
        self._node = node
        self._result: list[Operation] = []
        self._visit_expression(self.expression)
        if node.type == NodeType.RETURN:
            r = Return(get(self.expression))
            r.set_expression(expression)
            self._result.append(r)
        for ir in self._result:
            ir.set_node(node)

    def result(self) -> list[Operation]:
        return self._result

    def _post_assignement_operation(self, expression: AssignmentOperation) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        operation: Operation
        if isinstance(left, list):  # tuple expression:
            if isinstance(right, list):  # unbox assignment
                assert len(left) == len(right)
                for idx, _ in enumerate(left):
                    if (
                        left[idx] is not None
                        and expression.type
                        and expression.expression_return_type
                    ):
                        operation = convert_assignment(
                            left[idx],
                            right[idx],
                            expression.type,
                            expression.expression_return_type,
                        )
                        operation.set_expression(expression)
                        self._result.append(operation)
                set_val(expression, None)
            else:
                assert isinstance(right, TupleVariable)
                for idx, _ in enumerate(left):
                    if left[idx] is not None:
                        index = idx
                        operation = Unpack(left[idx], right, index)
                        operation.set_expression(expression)
                        self._result.append(operation)
                set_val(expression, None)
        # Tuple with only one element. We need to convert the assignment to a Unpack
        # Ex:
        # (uint a,,) = g()
        elif (
            isinstance(left, LocalVariableInitFromTuple)
            and left.tuple_index is not None
            and isinstance(right, TupleVariable)
        ):
            operation = Unpack(left, right, left.tuple_index)
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, None)
        else:
            # For `InitArray`, the rhs is a list or singleton of `TupleExpression` elements.
            # Init of array e.g. uint8[2] var = [1,2];
            if isinstance(right, list):
                operation = InitArray(right, left)
                operation.set_expression(expression)
                self._result.append(operation)
                set_val(expression, left)

            # Special case for init of array, when the right has only one element e.g. arr = [1];
            elif isinstance(left.type, ArrayType) and not isinstance(right.type, ArrayType):
                operation = InitArray([right], left)
                operation.set_expression(expression)
                self._result.append(operation)
                set_val(expression, left)

            elif (
                isinstance(left.type, UserDefinedType)
                and isinstance(left.type.type, Structure)
                and isinstance(right, TupleVariable)
            ):
                # This will result in a `NewStructure` operation where
                # each field is assigned the value unpacked from the tuple
                # (see `slither.vyper_parsing.type_parsing.parse_type`)
                args = []
                for idx, elem in enumerate(left.type.type.elems.values()):
                    temp = TemporaryVariable(self._node)
                    temp.type = elem.type
                    args.append(temp)
                    operation = Unpack(temp, right, idx)
                    operation.set_expression(expression)
                    self._result.append(operation)

                for arg in args:
                    op = Argument(arg)
                    op.set_expression(expression)
                    self._result.append(op)

                operation = TmpCall(
                    left.type.type, len(left.type.type.elems), left, left.type.type.name
                )
                operation.set_expression(expression)
                self._result.append(operation)

            else:
                operation = convert_assignment(
                    left, right, expression.type, expression.expression_return_type
                )
                operation.set_expression(expression)
                self._result.append(operation)
                # Return left to handle
                # a = b = 1;
                set_val(expression, left)

    def _post_binary_operation(self, expression: BinaryOperation) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        val = TemporaryVariable(self._node)

        if expression.type in _signed_to_unsigned:
            new_left = TemporaryVariable(self._node)
            conv_left = TypeConversion(new_left, left, ElementaryType("int256"))
            new_left.set_type(ElementaryType("int256"))
            conv_left.set_expression(expression)
            self._result.append(conv_left)

            if expression.type != BinaryOperationType.RIGHT_SHIFT_ARITHMETIC:
                new_right = TemporaryVariable(self._node)
                conv_right = TypeConversion(new_right, right, ElementaryType("int256"))
                new_right.set_type(ElementaryType("int256"))
                conv_right.set_expression(expression)
                self._result.append(conv_right)
            else:
                new_right = right

            new_final = TemporaryVariable(self._node)
            operation = Binary(new_final, new_left, new_right, _signed_to_unsigned[expression.type])
            operation.set_expression(expression)
            self._result.append(operation)

            conv_final = TypeConversion(val, new_final, ElementaryType("uint256"))
            val.set_type(ElementaryType("uint256"))
            conv_final.set_expression(expression)
            self._result.append(conv_final)
        else:
            operation = Binary(val, left, right, _binary_to_binary[expression.type])
            operation.set_expression(expression)
            self._result.append(operation)

        set_val(expression, val)

    def _post_call_expression(self, expression: CallExpression) -> None:
        assert isinstance(expression, CallExpression)

        expression_called = expression.called
        called = get(expression_called)

        args = [get(a) for a in expression.arguments if a]
        val: TupleVariable | TemporaryVariable
        var: Operation
        for arg in args:
            arg_ = Argument(arg)
            arg_.set_expression(expression)
            self._result.append(arg_)
        if isinstance(called, Function):
            # internal call

            # If tuple

            if expression.type_call.startswith("tuple(") and expression.type_call != "tuple()":
                val = TupleVariable(self._node)
            else:
                val = TemporaryVariable(self._node)
            internal_call = InternalCall(
                called, len(args), val, expression.type_call, names=expression.names
            )
            internal_call.set_expression(expression)
            self._result.append(internal_call)
            set_val(expression, val)

        # User defined types
        elif (
            isinstance(called, TypeAlias)
            and isinstance(expression_called, MemberAccess)
            and expression_called.member_name in ["wrap", "unwrap"]
            and len(args) == 1
        ):
            # wrap: underlying_type -> alias
            # unwrap: alias -> underlying_type
            dest_type: TypeAlias | ElementaryType = (
                called if expression_called.member_name == "wrap" else called.underlying_type
            )
            val = TemporaryVariable(self._node)
            var = TypeConversion(val, args[0], dest_type)
            var.set_expression(expression)
            val.set_type(dest_type)
            self._result.append(var)
            set_val(expression, val)

        # yul things
        elif called.name == "caller()":
            val = TemporaryVariable(self._node)
            var = Assignment(val, SolidityVariableComposed("msg.sender"), ElementaryType("uint256"))
            self._result.append(var)
            set_val(expression, val)
        elif called.name == "origin()":
            val = TemporaryVariable(self._node)
            var = Assignment(val, SolidityVariableComposed("tx.origin"), ElementaryType("uint256"))
            self._result.append(var)
            set_val(expression, val)
        elif called.name == "extcodesize(uint256)":
            val_ref = ReferenceVariable(self._node)
            var = Member(args[0], Constant("codesize"), val_ref)
            self._result.append(var)
            set_val(expression, val_ref)
        elif called.name == "selfbalance()":
            val = TemporaryVariable(self._node)
            var = TypeConversion(val, SolidityVariable("this"), ElementaryType("address"))
            val.set_type(ElementaryType("address"))
            self._result.append(var)

            val1 = ReferenceVariable(self._node)
            var1 = Member(val, Constant("balance"), val1)
            self._result.append(var1)
            set_val(expression, val1)
        elif called.name == "address()":
            val = TemporaryVariable(self._node)
            var = TypeConversion(val, SolidityVariable("this"), ElementaryType("address"))
            val.set_type(ElementaryType("address"))
            self._result.append(var)
            set_val(expression, val)
        elif called.name == "callvalue()":
            val = TemporaryVariable(self._node)
            var = Assignment(val, SolidityVariableComposed("msg.value"), ElementaryType("uint256"))
            self._result.append(var)
            set_val(expression, val)

        elif (
            called.name in ["sload(uint256)", "sstore(uint256,uint256)"]
            and len(args) > 0
            and isinstance(args[0], StateVariable)
        ):
            # parse_yul._parse_yul_magic_suffixes does a best effort tentative to retrieve
            # the right state variable on .slot access
            #
            # Solidity does not allow state variable to be directly used through sstore/sload
            # As you need to specify the slot number (ex you can't do " sload(some_state_variable)")
            #
            # So we can make the assumption that if a state variable appear on the first argument
            # of sstore/sload, we can convert the call to sstore to a normal assignment / read

            if called.name == "sload(uint256)":
                val = TemporaryVariable(self._node)
                var = Assignment(val, args[0], ElementaryType("uint256"))
                self._result.append(var)
                set_val(expression, val)
            else:
                var = Assignment(args[0], args[1], ElementaryType("uint256"))
                self._result.append(var)
                set_val(expression, args[0])
        else:
            # If tuple
            if expression.type_call.startswith("tuple(") and expression.type_call != "tuple()":
                val = TupleVariable(self._node)
            else:
                val = TemporaryVariable(self._node)

            message_call = TmpCall(
                called, len(args), val, expression.type_call, names=expression.names
            )
            message_call.set_expression(expression)
            # Gas/value are only accessible here if the syntax {gas: , value: }
            # Is used over .gas().value()
            if expression.call_gas:
                call_gas = get(expression.call_gas)
                message_call.call_gas = call_gas
            if expression.call_value:
                call_value = get(expression.call_value)
                message_call.call_value = call_value
            if expression.call_salt:
                call_salt = get(expression.call_salt)
                message_call.call_salt = call_salt
            self._result.append(message_call)
            set_val(expression, val)

    def _post_conditional_expression(self, expression: ConditionalExpression) -> None:
        raise SlithIRError(f"Ternary operator are not convertible to SlithIR {expression}")

    def _post_elementary_type_name_expression(
        self,
        expression: ElementaryTypeNameExpression,
    ) -> None:
        set_val(expression, expression.type)

    def _post_identifier(self, expression: Identifier) -> None:
        set_val(expression, expression.value)

    def _post_index_access(self, expression: IndexAccess) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        operation: Operation
        # Left can be a type for abi.decode(var, uint[2])
        if isinstance(left, (Type, Contract, Enum, Structure)):
            # Nested type are not yet supported by abi.decode, so the assumption
            # Is that the right variable must be a constant
            assert isinstance(right, Constant)
            # Case for abi.decode(var, I[2]) where I is an interface/contract or an enum
            if isinstance(left, (Contract, Enum, Structure)):
                left = UserDefinedType(left)
            t = ArrayType(left, int(right.value))
            set_val(expression, t)
            return
        val = ReferenceVariable(self._node)

        if (
            isinstance(left, LocalVariable)
            and isinstance(left.type, UserDefinedType)
            and isinstance(left.type.type, Structure)
        ):
            # We rewrite the index access to a tuple variable as
            # an access to its field i.e. the 0th element is the field "_0"
            # (see `slither.vyper_parsing.type_parsing.parse_type`)
            operation = Member(left, Constant("_" + str(right)), val)
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, val)
            return

        # access to anonymous array
        # such as [0,1][x]
        if isinstance(left, list):
            init_array_val = TemporaryVariable(self._node)
            init_array_right = left
            left = init_array_val
            operation = InitArray(init_array_right, init_array_val)
            operation.set_expression(expression)
            self._result.append(operation)

        operation = Index(val, left, right)
        operation.set_expression(expression)
        self._result.append(operation)
        set_val(expression, val)

    def _post_literal(self, expression: Literal) -> None:
        expression_type = expression.type
        assert isinstance(expression_type, ElementaryType)
        cst = Constant(expression.value, expression_type, expression.subdenomination)
        set_val(expression, cst)

    def _post_member_access(self, expression: MemberAccess) -> None:
        expr = get(expression.expression)

        # Look for type(X).max / min
        # Because we looked at the AST structure, we need to look into the nested expression
        # Hopefully this is always on a direct sub field, and there is no weird construction

        if isinstance(expression.expression, CallExpression) and expression.member_name in [
            "min",
            "max",
        ]:
            if isinstance(expression.expression.called, Identifier):
                if expression.expression.called.value == SolidityFunction("type()"):
                    assert len(expression.expression.arguments) == 1
                    val = TemporaryVariable(self._node)
                    type_expression_found = expression.expression.arguments[0]
                    type_found: ElementaryType | UserDefinedType
                    if isinstance(type_expression_found, ElementaryTypeNameExpression):
                        type_expression_found_type = type_expression_found.type
                        assert isinstance(type_expression_found_type, ElementaryType)
                        type_found = type_expression_found_type
                        min_value = type_found.min
                        max_value = type_found.max
                        constant_type = type_found
                    else:
                        # type(enum).max/min
                        # Case when enum is in another contract e.g. type(C.E).max
                        if isinstance(type_expression_found, MemberAccess):
                            contract = type_expression_found.expression.value
                            assert isinstance(contract, Contract)
                            for enum in contract.enums:
                                if enum.name == type_expression_found.member_name:
                                    type_found_in_expression = enum
                                    type_found = UserDefinedType(enum)
                                    break
                        else:
                            assert isinstance(type_expression_found, Identifier)
                            type_found_in_expression = type_expression_found.value
                            assert isinstance(
                                type_found_in_expression, (EnumContract, EnumTopLevel)
                            )
                            type_found = UserDefinedType(type_found_in_expression)
                        constant_type = None
                        min_value = type_found_in_expression.min
                        max_value = type_found_in_expression.max
                    if expression.member_name == "min":
                        op = Assignment(
                            val,
                            Constant(str(min_value), constant_type),
                            type_found,
                        )
                    else:
                        op = Assignment(
                            val,
                            Constant(str(max_value), constant_type),
                            type_found,
                        )
                    self._result.append(op)
                    set_val(expression, val)
                    return

        # This does not support solidity 0.4 contract_name.balance
        if (
            isinstance(expr, Variable)
            and expr.type == ElementaryType("address")
            and expression.member_name in ["balance", "code", "codehash"]
        ):
            val = TemporaryVariable(self._node)
            name = expression.member_name + "(address)"
            sol_func = SolidityFunction(name)
            s = SolidityCall(
                sol_func,
                1,
                val,
                sol_func.return_type,
            )
            s.set_expression(expression)
            s.arguments.append(expr)
            self._result.append(s)
            set_val(expression, val)
            return

        if isinstance(expr, TypeAlias) and expression.member_name in ["wrap", "unwrap"]:
            # The logic is be handled by _post_call_expression
            set_val(expression, expr)
            return

        if isinstance(expr, Contract):
            # Early lookup to detect user defined types from other contracts definitions
            # contract A { type MyInt is int}
            # contract B { function f() public{ A.MyInt test = A.MyInt.wrap(1);}}
            # The logic is handled by _post_call_expression
            if expression.member_name in expr.type_aliases_as_dict:
                set_val(expression, expr.type_aliases_as_dict[expression.member_name])
                return
            # Lookup errors referred to as member of contract e.g. Test.myError.selector
            if expression.member_name in expr.custom_errors_as_dict:
                set_val(expression, expr.custom_errors_as_dict[expression.member_name])
                return
            # Lookup enums when in a different contract e.g. C.E
            if str(expression) in expr.enums_as_dict:
                set_val(expression, expr.enums_as_dict[str(expression)])
                return

        if isinstance(expr, (SolidityImportPlaceHolder, Import)):
            scope = (
                expr.import_directive.scope
                if isinstance(expr, SolidityImportPlaceHolder)
                else expr.scope
            )
            if self._check_elem_in_scope(expression.member_name, scope, expression):
                return

        val_ref = ReferenceVariable(self._node)
        member = Member(expr, Constant(expression.member_name), val_ref)
        member.set_expression(expression)
        self._result.append(member)
        set_val(expression, val_ref)

    def _check_elem_in_scope(self, elem: str, scope: FileScope, expression: MemberAccess) -> bool:
        if elem in scope.renaming:
            self._check_elem_in_scope(scope.renaming[elem], scope, expression)
            return True

        if elem in scope.contracts:
            set_val(expression, scope.contracts[elem])
            return True

        if elem in scope.structures:
            set_val(expression, scope.structures[elem])
            return True

        if elem in scope.variables:
            set_val(expression, scope.variables[elem])
            return True

        if elem in scope.enums:
            set_val(expression, scope.enums[elem])
            return True

        if elem in scope.type_aliases:
            set_val(expression, scope.type_aliases[elem])
            return True

        for import_directive in scope.imports:
            if elem == import_directive.alias:
                set_val(expression, import_directive)
                return True

        for custom_error in scope.custom_errors:
            if custom_error.name == elem:
                set_val(expression, custom_error)
                return True

        if str(expression.type).startswith("function "):
            # This is needed to handle functions overloading
            signature_to_seaarch = (
                str(expression.type)
                .replace("function ", elem)
                .replace("pure ", "")
                .replace("view ", "")
                .replace("struct ", "")
                .replace("enum ", "")
                .replace(" memory", "")
                .split(" returns", maxsplit=1)[0]
            )

            for function in scope.functions:
                if signature_to_seaarch == function.full_name:
                    set_val(expression, function)
                    return True

        return False

    def _post_new_array(self, expression: NewArray) -> None:
        val = TemporaryVariable(self._node)
        operation = TmpNewArray(expression.array_type, val)
        operation.set_expression(expression)
        self._result.append(operation)
        set_val(expression, val)

    def _post_new_contract(self, expression: NewContract) -> None:
        val = TemporaryVariable(self._node)
        operation = TmpNewContract(expression.contract_name, val)
        operation.set_expression(expression)
        if expression.call_value:
            call_value = get(expression.call_value)
            operation.call_value = call_value
        if expression.call_salt:
            call_salt = get(expression.call_salt)
            operation.call_salt = call_salt

        self._result.append(operation)
        set_val(expression, val)

    def _post_new_elementary_type(self, expression: NewElementaryType) -> None:
        # TODO unclear if this is ever used?
        val = TemporaryVariable(self._node)
        operation = TmpNewElementaryType(expression.type, val)
        operation.set_expression(expression)
        self._result.append(operation)
        set_val(expression, val)

    def _post_tuple_expression(self, expression: TupleExpression) -> None:
        all_expressions = [get(e) if e else None for e in expression.expressions]
        if len(all_expressions) == 1:
            val = all_expressions[0]
        else:
            val = all_expressions
        set_val(expression, val)

    def _post_type_conversion(self, expression: expressions.TypeConversion) -> None:
        assert expression.expression
        expr = get(expression.expression)
        val = TemporaryVariable(self._node)
        expression_type = expression.type
        assert isinstance(expression_type, (TypeAlias, UserDefinedType, ElementaryType, ArrayType))
        operation = TypeConversion(val, expr, expression_type)
        val.set_type(expression.type)
        operation.set_expression(expression)
        self._result.append(operation)
        set_val(expression, val)

    def _post_unary_operation(self, expression: UnaryOperation) -> None:
        value = get(expression.expression)
        operation: Operation
        if expression.type in [UnaryOperationType.BANG, UnaryOperationType.TILD]:
            lvalue = TemporaryVariable(self._node)
            operation = Unary(lvalue, value, _unary_to_unary[expression.type])
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, lvalue)
        elif expression.type in [UnaryOperationType.DELETE]:
            operation = Delete(value, value)
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, value)
        elif expression.type in [UnaryOperationType.PLUSPLUS_PRE]:
            operation = Binary(value, value, Constant("1", value.type), BinaryType.ADDITION)
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, value)
        elif expression.type in [UnaryOperationType.MINUSMINUS_PRE]:
            operation = Binary(value, value, Constant("1", value.type), BinaryType.SUBTRACTION)
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, value)
        elif expression.type in [UnaryOperationType.PLUSPLUS_POST]:
            lvalue = TemporaryVariable(self._node)
            operation = Assignment(lvalue, value, value.type)
            operation.set_expression(expression)
            self._result.append(operation)
            operation = Binary(value, value, Constant("1", value.type), BinaryType.ADDITION)
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, lvalue)
        elif expression.type in [UnaryOperationType.MINUSMINUS_POST]:
            lvalue = TemporaryVariable(self._node)
            operation = Assignment(lvalue, value, value.type)
            operation.set_expression(expression)
            self._result.append(operation)
            operation = Binary(value, value, Constant("1", value.type), BinaryType.SUBTRACTION)
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, lvalue)
        elif expression.type in [UnaryOperationType.PLUS_PRE]:
            set_val(expression, value)
        elif expression.type in [UnaryOperationType.MINUS_PRE]:
            lvalue = TemporaryVariable(self._node)
            operation = Binary(lvalue, Constant("0", value.type), value, BinaryType.SUBTRACTION)
            operation.set_expression(expression)
            self._result.append(operation)
            set_val(expression, lvalue)
        else:
            raise SlithIRError(f"Unary operation to IR not supported {expression}")



"""
Module printing summary of the contract
"""

from slither.core.declarations import Function
from slither.core.declarations.function import SolidityFunction
from slither.printers.abstract_printer import AbstractPrinter
from slither.utils import output
from slither.utils.myprettytable import MyPrettyTable


def _use_modifier(function: Function, modifier_name: str = "whenNotPaused") -> bool:
    for ir in function.all_internal_calls():
        if isinstance(ir, function, SolidityFunction):
            continue
        if any(modifier.name == modifier_name for modifier in function.modifiers):
            return True
    return False


class PrinterWhenNotPaused(AbstractPrinter):
    ARGUMENT = "not-pausable"
    HELP = "Print functions that do not use whenNotPaused"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#when-not-paused"

    def output(self, _filename: str) -> output.Output:
        """
        _filename is not used
        Args:
            _filename(string)
        """

        modifier_name: str = "whenNotPaused"

        txt = ""
        txt += "Constructor and pure/view functions are not displayed\n"
        all_tables = []
        for contract in self.slither.contracts:
            txt += f"\n{contract.name}:\n"
            table = MyPrettyTable(["Name", "Use whenNotPaused"])

            for function in contract.functions_entry_points:
                if function.is_constructor or function.view or function.pure:
                    continue
                status = "X" if _use_modifier(function, modifier_name) else ""
                table.add_row([function.solidity_signature, status])

            txt += str(table) + "\n"
            all_tables.append((contract.name, table))

        self.info(txt)

        res = self.generate_output(txt)
        for name, table in all_tables:
            res.add_pretty_table(table, name)

        return res



from typing import Union, TYPE_CHECKING

from slither.core.expressions.expression import Expression
from slither.core.solidity_types.type import Type

if TYPE_CHECKING:
    from slither.core.expressions.call_expression import CallExpression
    from slither.core.expressions.identifier import Identifier
    from slither.core.expressions.literal import Literal
    from slither.core.expressions.member_access import MemberAccess
    from slither.core.solidity_types.elementary_type import ElementaryType
    from slither.core.solidity_types.type_alias import TypeAliasContract
    from slither.core.solidity_types.user_defined_type import UserDefinedType


class TypeConversion(Expression):
    def __init__(
        self,
        expression: Union[
            "MemberAccess", "Literal", "CallExpression", "TypeConversion", "Identifier"
        ],
        expression_type: Union["ElementaryType", "UserDefinedType", "TypeAliasContract"],
    ) -> None:
        super().__init__()
        assert isinstance(expression, Expression)
        assert isinstance(expression_type, Type)
        self._expression: Expression = expression
        self._type: Type = expression_type

    @property
    def type(self) -> Type:
        return self._type

    @type.setter
    def type(self, new_type: Type) -> None:
        self._type = new_type

    @property
    def expression(self) -> Expression:
        return self._expression

    def __str__(self) -> str:
        return str(self.type) + "(" + str(self.expression) + ")"



from slither.core.declarations.structure import Structure
from slither.core.variables.structure_variable import StructureVariable
from slither.vyper_parsing.variables.structure_variable import StructureVariableVyper
from slither.vyper_parsing.ast.types import StructDef, AnnAssign


class StructVyper:
    def __init__(
        self,
        st: Structure,
        struct: StructDef,
    ) -> None:
        self._structure = st
        st.name = struct.name
        st.canonical_name = struct.name + self._structure.contract.name

        self._elemsNotParsed: list[AnnAssign] = struct.body

    def analyze(self, contract) -> None:
        for elem_to_parse in self._elemsNotParsed:
            elem = StructureVariable()
            elem.set_structure(self._structure)
            elem.set_offset(elem_to_parse.src, self._structure.contract.compilation_unit)

            elem_parser = StructureVariableVyper(elem, elem_to_parse)
            elem_parser.analyze(contract)

            self._structure.elems[elem.name] = elem
            self._structure.add_elem_in_order(elem.name)
        self._elemsNotParsed = []



import logging
from enum import Enum

from slither.core.expressions.expression import Expression
from slither.core.exceptions import SlitherCoreError
from slither.core.expressions.identifier import Identifier
from slither.core.expressions.index_access import IndexAccess
from slither.core.expressions.literal import Literal
from slither.core.expressions.tuple_expression import TupleExpression


logger = logging.getLogger("UnaryOperation")


class UnaryOperationType(Enum):
    BANG = 0  # !
    TILD = 1  # ~
    DELETE = 2  # delete
    PLUSPLUS_PRE = 3  # ++
    MINUSMINUS_PRE = 4  # --
    PLUSPLUS_POST = 5  # ++
    MINUSMINUS_POST = 6  # --
    PLUS_PRE = 7  # for stuff like uint(+1)
    MINUS_PRE = 8  # for stuff like uint(-1)

    @staticmethod
    def get_type(operation_type: str, isprefix: bool) -> "UnaryOperationType":
        if isprefix:
            if operation_type == "!":
                return UnaryOperationType.BANG
            if operation_type == "~":
                return UnaryOperationType.TILD
            if operation_type == "delete":
                return UnaryOperationType.DELETE
            if operation_type == "++":
                return UnaryOperationType.PLUSPLUS_PRE
            if operation_type == "--":
                return UnaryOperationType.MINUSMINUS_PRE
            if operation_type == "+":
                return UnaryOperationType.PLUS_PRE
            if operation_type == "-":
                return UnaryOperationType.MINUS_PRE
        else:
            if operation_type == "++":
                return UnaryOperationType.PLUSPLUS_POST
            if operation_type == "--":
                return UnaryOperationType.MINUSMINUS_POST
        raise SlitherCoreError(f"get_type: Unknown operation type {operation_type}")

    def __str__(self) -> str:
        if self == UnaryOperationType.BANG:
            return "!"
        if self == UnaryOperationType.TILD:
            return "~"
        if self == UnaryOperationType.DELETE:
            return "delete"
        if self == UnaryOperationType.PLUS_PRE:
            return "+"
        if self == UnaryOperationType.MINUS_PRE:
            return "-"
        if self in [UnaryOperationType.PLUSPLUS_PRE, UnaryOperationType.PLUSPLUS_POST]:
            return "++"
        if self in [
            UnaryOperationType.MINUSMINUS_PRE,
            UnaryOperationType.MINUSMINUS_POST,
        ]:
            return "--"

        raise SlitherCoreError(f"str: Unknown operation type {self}")

    @staticmethod
    def is_prefix(operation_type: "UnaryOperationType") -> bool:
        if operation_type in [
            UnaryOperationType.BANG,
            UnaryOperationType.TILD,
            UnaryOperationType.DELETE,
            UnaryOperationType.PLUSPLUS_PRE,
            UnaryOperationType.MINUSMINUS_PRE,
            UnaryOperationType.PLUS_PRE,
            UnaryOperationType.MINUS_PRE,
        ]:
            return True
        if operation_type in [
            UnaryOperationType.PLUSPLUS_POST,
            UnaryOperationType.MINUSMINUS_POST,
        ]:
            return False

        raise SlitherCoreError(f"is_prefix: Unknown operation type {operation_type}")


class UnaryOperation(Expression):
    def __init__(
        self,
        expression: Literal | Identifier | IndexAccess | TupleExpression,
        expression_type: UnaryOperationType,
    ) -> None:
        assert isinstance(expression, Expression)
        super().__init__()
        self._expression: Expression = expression
        self._type: UnaryOperationType = expression_type
        if expression_type in [
            UnaryOperationType.DELETE,
            UnaryOperationType.PLUSPLUS_PRE,
            UnaryOperationType.MINUSMINUS_PRE,
            UnaryOperationType.PLUSPLUS_POST,
            UnaryOperationType.MINUSMINUS_POST,
        ]:
            expression.set_lvalue()

    @property
    def expression(self) -> Expression:
        return self._expression

    @property
    def type(self) -> UnaryOperationType:
        return self._type

    @property
    def is_prefix(self) -> bool:
        return UnaryOperationType.is_prefix(self._type)

    def __str__(self) -> str:
        if self.is_prefix:
            return str(self.type) + " " + str(self._expression)
        return str(self._expression) + " " + str(self.type)



from slither.core.expressions.expression import Expression
from slither.visitors.expression.expression import ExpressionVisitor
from slither.core.expressions.conditional_expression import ConditionalExpression


class HasConditional(ExpressionVisitor):
    def __init__(self, expression: Expression) -> None:
        self._result: bool = False
        super().__init__(expression)

    def result(self) -> bool:
        return self._result

    def _post_conditional_expression(self, expression: ConditionalExpression) -> None:
        self._result = True



import logging
from collections import namedtuple
from pathlib import Path

from crytic_compile.platform.abstract_platform import AbstractPlatform
from crytic_compile.platform import Type as PlatformType

from slither.core.declarations import Contract
from slither.tools.properties.addresses.address import Addresses
from slither.tools.properties.platforms.echidna import generate_echidna_config
from slither.tools.properties.properties.ercs.erc20.properties.burn import (
    ERC20_NotBurnable,
    ERC20_Burnable,
)
from slither.tools.properties.properties.ercs.erc20.properties.initialization import ERC20_CONFIG
from slither.tools.properties.properties.ercs.erc20.properties.mint import ERC20_NotMintable
from slither.tools.properties.properties.ercs.erc20.properties.mint_and_burn import (
    ERC20_NotMintableNotBurnable,
)
from slither.tools.properties.properties.ercs.erc20.properties.transfer import (
    ERC20_Transferable,
    ERC20_Pausable,
)
from slither.tools.properties.properties.ercs.erc20.unit_tests.truffle import generate_truffle_test
from slither.tools.properties.properties.properties import (
    property_to_solidity,
    Property,
)
from slither.tools.properties.solidity.generate_properties import (
    generate_solidity_properties,
    generate_test_contract,
    generate_solidity_interface,
)
from slither.utils.colors import red, green

logger = logging.getLogger("Slither")

PropertyDescription = namedtuple("PropertyDescription", ["properties", "description"])

ERC20_PROPERTIES = {
    "Transferable": PropertyDescription(ERC20_Transferable, "Test the correct tokens transfer"),
    "Pausable": PropertyDescription(ERC20_Pausable, "Test the pausable functionality"),
    "NotMintable": PropertyDescription(ERC20_NotMintable, "Test that no one can mint tokens"),
    "NotMintableNotBurnable": PropertyDescription(
        ERC20_NotMintableNotBurnable, "Test that no one can mint or burn tokens"
    ),
    "NotBurnable": PropertyDescription(ERC20_NotBurnable, "Test that no one can burn tokens"),
    "Burnable": PropertyDescription(
        ERC20_Burnable,
        'Test the burn of tokens. Require the "burn(address) returns()" function',
    ),
}


def generate_erc20(contract: Contract, type_property: str, addresses: Addresses):
    """
    Generate the ERC20 tests
    Files generated:
    - interfaces.sol: generic crytic interface
    - Properties[CONTRACTNAME].sol: erc20 properties
    - Test[CONTRACTNAME].sol: Target, its constructor needs to be manually updated
    - If truffle
        - migrations/x_Test[CONTRACTNAME].js
        - test/crytic/InitializationTest[CONTRACTNAME].js: unit tests to check that the contract is correctly configured
        - test/crytic/Test[CONTRACTNAME].js: ERC20 checks
    - echidna_config.yaml: configuration file
    :param addresses:
    :param contract:
    :param type_property: One of ERC20_PROPERTIES.keys()
    :return:
    """
    if contract.compilation_unit.core.crytic_compile is None:
        logging.error("Please compile with crytic-compile")
        return
    if contract.compilation_unit.core.crytic_compile.type not in [
        PlatformType.TRUFFLE,
        PlatformType.SOLC,
        PlatformType.BUILDER,
    ]:
        logging.error(
            f"{contract.compilation_unit.core.crytic_compile.type} not yet supported by slither-prop"
        )
        return

    # Check if the contract is an ERC20 contract and if the functions have the correct visibility
    errors = _check_compatibility(contract)
    if errors:
        logger.error(red(errors))
        return

    erc_properties = ERC20_PROPERTIES.get(type_property)
    if erc_properties is None:
        logger.error(f"{type_property} unknown. Types available {ERC20_PROPERTIES.keys()}")
        return
    properties = erc_properties.properties

    # Generate the output directory
    output_dir = _platform_to_output_dir(contract.compilation_unit.core.crytic_compile.platform)
    output_dir.mkdir(exist_ok=True)

    # Get the properties
    solidity_properties, unit_tests = _get_properties(contract, properties)

    # Generate the contract containing the properties
    generate_solidity_interface(output_dir, addresses)
    property_file = generate_solidity_properties(
        contract, type_property, solidity_properties, output_dir
    )

    # Generate the Test contract
    initialization_recommendation = _initialization_recommendation(type_property)
    contract_filename, contract_name = generate_test_contract(
        contract,
        type_property,
        output_dir,
        property_file,
        initialization_recommendation,
    )

    # Generate Echidna config file
    echidna_config_filename = generate_echidna_config(
        Path(contract.compilation_unit.core.crytic_compile.target).parent, addresses
    )

    unit_test_info = ""

    # If truffle, generate unit tests
    if contract.compilation_unit.core.crytic_compile.type == PlatformType.TRUFFLE:
        unit_test_info = generate_truffle_test(contract, type_property, unit_tests, addresses)

    logger.info("################################################")
    logger.info(green(f"Update the constructor in {Path(output_dir, contract_filename)}"))

    if unit_test_info:
        logger.info(green(unit_test_info))

    logger.info(green("To run Echidna:"))
    txt = f"\t echidna-test {contract.compilation_unit.core.crytic_compile.target} "
    txt += f"--contract {contract_name} --config {echidna_config_filename}"
    logger.info(green(txt))


def _initialization_recommendation(type_property: str) -> str:
    content = ""
    content += "\t\t// Add below a minimal configuration:\n"
    content += "\t\t// - crytic_owner must have some tokens \n"
    content += "\t\t// - crytic_user must have some tokens \n"
    content += "\t\t// - crytic_attacker must have some tokens \n"
    if type_property in ["Pausable"]:
        content += "\t\t// - The contract must be paused \n"
    if type_property in ["NotMintable", "NotMintableNotBurnable"]:
        content += "\t\t// - The contract must not be mintable \n"
    if type_property in ["NotBurnable", "NotMintableNotBurnable"]:
        content += "\t\t// - The contract must not be burnable \n"
    content += "\n"
    content += "\n"

    return content


# TODO: move this to crytic-compile
def _platform_to_output_dir(platform: AbstractPlatform) -> Path:
    if platform.TYPE in [PlatformType.TRUFFLE, platform.TYPE == PlatformType.BUILDER]:
        return Path(platform.target, "contracts", "crytic")
    if platform.TYPE == PlatformType.SOLC:
        return Path(platform.target).parent
    return Path()


def _check_compatibility(contract):
    errors = ""
    if not contract.is_erc20():
        errors = f"{contract} is not ERC20 compliant. Consider checking the contract with slither-check-erc"
        return errors

    transfer = contract.get_function_from_signature("transfer(address,uint256)")

    if transfer.visibility != "public":
        errors = f"slither-prop requires {transfer.canonical_name} to be public. Please change the visibility"

    transfer_from = contract.get_function_from_signature("transferFrom(address,address,uint256)")
    if transfer_from.visibility != "public":
        if errors:
            errors += "\n"
        errors += f"slither-prop requires {transfer_from.canonical_name} to be public. Please change the visibility"

    approve = contract.get_function_from_signature("approve(address,uint256)")
    if approve.visibility != "public":
        if errors:
            errors += "\n"
        errors += f"slither-prop requires {approve.canonical_name} to be public. Please change the visibility"

    return errors


def _get_properties(contract, properties: list[Property]) -> tuple[str, list[Property]]:
    solidity_properties = ""

    if contract.compilation_unit.crytic_compile.type == PlatformType.TRUFFLE:
        solidity_properties += "\n".join([property_to_solidity(p) for p in ERC20_CONFIG])

    solidity_properties += "\n".join([property_to_solidity(p) for p in properties])
    unit_tests = [p for p in properties if p.is_unit_test]

    return solidity_properties, unit_tests



from slither.slithir.operations import Binary, BinaryType
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator

logical_operators = [
    BinaryType.OROR,
    BinaryType.ANDAND,
]


class LOR(AbstractMutator):
    NAME = "LOR"
    HELP = "Logical Operator Replacement"

    def _mutate(self) -> dict:
        result: dict = {}

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                for ir in node.irs:
                    if isinstance(ir, Binary) and ir.type in logical_operators:
                        alternative_ops = logical_operators[:]
                        alternative_ops.remove(ir.type)

                        for op in alternative_ops:
                            # Get the string
                            start = node.source_mapping.start
                            stop = start + node.source_mapping.length
                            old_str = node.source_mapping.content
                            line_no = node.source_mapping.lines
                            # Replace the expression with true
                            new_str = f"{old_str.split(ir.type.value)[0]} {op.value} {old_str.split(ir.type.value)[1]}"
                            create_patch_with_line(
                                result,
                                self.in_file,
                                start,
                                stop,
                                old_str,
                                new_str,
                                line_no[0],
                            )
        return result



from slither.tools.upgradeability.checks.abstract_checks import (
    CheckClassification,
    AbstractCheck,
    CHECK_INFO,
)
from slither.utils.output import Output


class VariableWithInit(AbstractCheck):
    ARGUMENT = "variables-initialized"
    IMPACT = CheckClassification.HIGH

    HELP = "State variables with an initial value"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#state-variable-initialized"
    WIKI_TITLE = "State variable initialized"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect state variables that are initialized.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    uint variable = 10;
}
```
Using `Contract` will the delegatecall proxy pattern will lead `variable` to be 0 when called through the proxy.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Using initialize functions to write initial values in state variables.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True

    def _check(self) -> list[Output]:
        results = []
        for s in self.contract.storage_variables_ordered:
            if s.initialized:
                info: CHECK_INFO = [s, " is a state variable with an initial value.\n"]
                json = self.generate_result(info)
                results.append(json)
        return results



import logging
from enum import Enum

from slither.core.declarations import Function
from slither.core.solidity_types import ElementaryType
from slither.core.variables.variable import Variable
from slither.slithir.exceptions import SlithIRError
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue, is_valid_rvalue, LVALUE, RVALUE
from slither.slithir.variables import ReferenceVariable

logger = logging.getLogger("BinaryOperationIR")


class BinaryType(Enum):
    POWER = "**"
    MULTIPLICATION = "*"
    DIVISION = "/"
    MODULO = "%"
    ADDITION = "+"
    SUBTRACTION = "-"
    LEFT_SHIFT = "<<"
    RIGHT_SHIFT = ">>"
    AND = "&"
    CARET = "^"
    OR = "|"
    LESS = "<"
    GREATER = ">"
    LESS_EQUAL = "<="
    GREATER_EQUAL = ">="
    EQUAL = "=="
    NOT_EQUAL = "!="
    ANDAND = "&&"
    OROR = "||"

    @staticmethod
    def return_bool(operation_type: "BinaryType") -> bool:
        return operation_type in [
            BinaryType.OROR,
            BinaryType.ANDAND,
            BinaryType.LESS,
            BinaryType.GREATER,
            BinaryType.LESS_EQUAL,
            BinaryType.GREATER_EQUAL,
            BinaryType.EQUAL,
            BinaryType.NOT_EQUAL,
        ]

    @staticmethod
    def get_type(operation_type: str) -> "BinaryType":
        if operation_type == "**":
            return BinaryType.POWER
        if operation_type == "*":
            return BinaryType.MULTIPLICATION
        if operation_type == "/":
            return BinaryType.DIVISION
        if operation_type == "%":
            return BinaryType.MODULO
        if operation_type == "+":
            return BinaryType.ADDITION
        if operation_type == "-":
            return BinaryType.SUBTRACTION
        if operation_type == "<<":
            return BinaryType.LEFT_SHIFT
        if operation_type == ">>":
            return BinaryType.RIGHT_SHIFT
        if operation_type == "&":
            return BinaryType.AND
        if operation_type == "^":
            return BinaryType.CARET
        if operation_type == "|":
            return BinaryType.OR
        if operation_type == "<":
            return BinaryType.LESS
        if operation_type == ">":
            return BinaryType.GREATER
        if operation_type == "<=":
            return BinaryType.LESS_EQUAL
        if operation_type == ">=":
            return BinaryType.GREATER_EQUAL
        if operation_type == "==":
            return BinaryType.EQUAL
        if operation_type == "!=":
            return BinaryType.NOT_EQUAL
        if operation_type == "&&":
            return BinaryType.ANDAND
        if operation_type == "||":
            return BinaryType.OROR

        raise SlithIRError(f"get_type: Unknown operation type {operation_type})")

    def can_be_checked_for_overflow(self) -> bool:
        return self in [
            BinaryType.POWER,
            BinaryType.MULTIPLICATION,
            BinaryType.ADDITION,
            BinaryType.SUBTRACTION,
            BinaryType.DIVISION,
        ]


class Binary(OperationWithLValue):
    def __init__(
        self,
        result: Variable,
        left_variable: RVALUE | Function,
        right_variable: RVALUE | Function,
        operation_type: BinaryType,
    ) -> None:
        assert is_valid_rvalue(left_variable) or isinstance(left_variable, Function)
        assert is_valid_rvalue(right_variable) or isinstance(right_variable, Function)
        assert is_valid_lvalue(result)
        assert isinstance(operation_type, BinaryType)
        super().__init__()
        self._variables = [left_variable, right_variable]
        self._type = operation_type
        self._lvalue = result
        if BinaryType.return_bool(operation_type):
            result.set_type(ElementaryType("bool"))
        else:
            result.set_type(left_variable.type)

    @property
    def read(self) -> list[RVALUE | LVALUE | Function]:
        return [self.variable_left, self.variable_right]

    @property
    def get_variable(self) -> list[RVALUE | Function]:
        return self._variables

    @property
    def variable_left(self) -> RVALUE | Function:
        return self._variables[0]  # type: ignore

    @property
    def variable_right(self) -> RVALUE | Function:
        return self._variables[1]  # type: ignore

    @property
    def type(self) -> BinaryType:
        return self._type

    @property
    def type_str(self) -> str:
        if self.node.scope.is_checked and self._type.can_be_checked_for_overflow():
            return "(c)" + str(self._type.value)
        return str(self._type.value)

    def __str__(self) -> str:
        lvalue = self.lvalue
        assert lvalue
        if isinstance(lvalue, ReferenceVariable):
            points = lvalue.points_to
            while isinstance(points, ReferenceVariable):
                points = points.points_to
            return f"{lvalue!s}(-> {points}) = {self.variable_left} {self.type_str} {self.variable_right}"

        return f"{lvalue!s}({lvalue.type}) = {self.variable_left} {self.type_str} {self.variable_right}"



from typing import Union, Any, TYPE_CHECKING

from slither.core.expressions.expression import Expression
from slither.core.expressions.literal import Literal
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.core.solidity_types.type import Type
from slither.visitors.expression.constants_folding import ConstantFolding

if TYPE_CHECKING:
    from slither.core.expressions.binary_operation import BinaryOperation
    from slither.core.expressions.identifier import Identifier


class ArrayType(Type):
    def __init__(
        self,
        t: Type,
        length: Union["Identifier", Literal, "BinaryOperation", int] | None,
    ) -> None:
        assert isinstance(t, Type)
        if length:
            if isinstance(length, int):
                length = Literal(length, ElementaryType("uint256"))

        super().__init__()
        self._type: Type = t
        assert length is None or isinstance(length, Expression)
        self._length: Expression | None = length

        if length:
            if not isinstance(length, Literal):
                cf = ConstantFolding(length, "uint256")
                length = cf.result()
            self._length_value: Literal | None = length
        else:
            self._length_value = None

    @property
    def type(self) -> Type:
        return self._type

    @property
    def is_dynamic(self) -> bool:
        return self.length is None

    @property
    def length(self) -> Expression | None:
        return self._length

    @property
    def length_value(self) -> Literal | None:
        return self._length_value

    @property
    def is_fixed_array(self) -> bool:
        return bool(self.length)

    @property
    def is_dynamic_array(self) -> bool:
        return not self.is_fixed_array

    @property
    def storage_size(self) -> tuple[int, bool]:
        if self._length_value:
            elem_size, _ = self._type.storage_size
            return elem_size * int(str(self._length_value)), True
        return 32, True

    def __str__(self) -> str:
        if self._length:
            return str(self._type) + f"[{self._length_value!s}]"
        return str(self._type) + "[]"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ArrayType):
            return False
        return self._type == other.type and self._length_value == other.length_value

    def __hash__(self) -> int:
        return hash(str(self))
